from . import _composition
from . import overrides
from . import variable_functions as _variable_functions
from ._composition import (
    Composition,
    IndexComposition,
    enable_c_accessing,
    set_c_accessing_enabled,
    is_c_accessing_enabled,
)
from . import core

from .core import decOVF


# TODO: need update
__all__ = [
    "Composition",
    "IndexComposition",
    "enable_c_accessing",
    "set_c_accessing_enabled",
    "is_c_accessing_enabled",
]

import typing as _typing
import torch.types as _types

PRIVATE_NAME = ["memory_format", "strided"]
PRIVATE_NAME.extend(dir(_typing))
PRIVATE_NAME.extend(dir(_types))


from typing import TYPE_CHECKING

# Fake import for type checking
if TYPE_CHECKING:
    from .variable_functions import *

for name in dir(_variable_functions):
    if name.startswith("__") or name in PRIVATE_NAME:
        continue
    obj = getattr(_variable_functions, name)
    obj.__module__ = "pydec"
    globals()[name] = obj
    if not name.startswith("_"):
        __all__.append(name)

import pydec.nn as nn



from __future__ import annotations

import pydec
import torch
from torch import Tensor

from typing import Dict, Tuple, Union, Any, Callable, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ..._composition import Composition

from torch.types import (
    _int,
    _float,
    _bool,
    Number,
    _dtype,
    _device,
    _qscheme,
    _size,
    _layout,
)

import pydec
from pydec.core.decOVF import register_decomposition_func, set_decomposition_func

__all__ = [
    "abs_affine_decomposition",
    "hybrid_affine_decomposition",
    "affine_decomposition",
    "scaling_decomposition",
    "_none_decomposition",
]


# TODO: all inplace operations should be re-examined, `ref` args should not be overridden
# TODO: algorithms should support `out` argument


@register_decomposition_func("abs_affine")
def abs_affine_decomposition(
    input: Composition,
    func: Callable[[Tensor], Tensor],
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
    inplace: _bool = False,
) -> Composition:
    if ref is None:
        recovery = input.c_sum()
    else:
        recovery = ref
    recovery_out = func(recovery)
    residual_out = func(input._residual_tensor)

    decompose_out = recovery_out - residual_out

    composition = input._component_tensor
    abs_composition = composition.abs()

    multiplier = decompose_out / abs_composition.sum(dim=0)

    if inplace:
        input._component_tensor.abs_()
        input._component_tensor *= multiplier
        input._residual_tensor = residual_out
        return input
    else:
        out_component_tensor = abs_composition * multiplier
        out_residual_tensor = residual_out
        return pydec.as_composition(out_component_tensor, out_residual_tensor)


@register_decomposition_func("hybrid_affine")
def hybrid_affine_decomposition(
    input: Composition,
    func: Callable[[Tensor], Tensor],
    *,
    ref: Optional[Tensor] = None,
    threshold: _float = 0,
    inplace: _bool = False,
) -> Composition:
    if ref is None:
        recovery = input.c_sum()
    else:
        recovery = ref
    recovery_out = func(recovery)
    residual_out = func(input._residual_tensor)

    decompose_out = recovery_out - residual_out

    composition = input._component_tensor
    if threshold > 0:
        sum_composition = composition.sum(dim=0)
        abs_composition = composition.abs()
        abs_sum_composition = abs_composition.sum(dim=0, keepdim=True)
        instability_ratio = sum_composition.abs() / abs_sum_composition
        mask = (instability_ratio < threshold).expand_as(composition)

        if not inplace:
            composition = composition.clone()

        composition[mask] = composition[mask].abs()

    multiplier = decompose_out / composition.sum(dim=0)
    multiplier.nan_to_num_(0, 0, 0)

    if inplace:
        input._component_tensor *= multiplier
        input._residual_tensor = residual_out
        return input
    else:
        out_component_tensor = composition * multiplier
        out_residual_tensor = residual_out
        return pydec.as_composition(out_component_tensor, out_residual_tensor)


@register_decomposition_func("affine")
def affine_decomposition(
    input: Composition,
    func: Callable[[Tensor], Tensor],
    *,
    ref: Optional[Tensor] = None,
    inplace: _bool = False,
) -> Composition:
    if ref is None:
        recovery = input.c_sum()
    else:
        recovery = ref
    recovery_out = func(recovery)
    residual_out = func(input._residual_tensor)

    decompose_out = recovery_out - residual_out

    composition = input._component_tensor
    multiplier = decompose_out / composition.sum(dim=0)

    # TODO: After processing NANs the residuals should be added to `out.residual`.
    # Also add a warning when there are NANs
    multiplier.nan_to_num_(0, 0, 0)

    if inplace:
        input._component_tensor *= multiplier
        input._residual_tensor = residual_out
        return input
    else:
        out_component_tensor = composition * multiplier
        out_residual_tensor = residual_out
        return pydec.as_composition(out_component_tensor, out_residual_tensor)


@register_decomposition_func("none")
def _none_decomposition(
    input: Composition,
    func: Callable[[Tensor], Tensor],
    *,
    ref: Optional[Tensor] = None,
    inplace: _bool = False,
) -> Composition:
    r"""
    Note: since PyDec 0.2.0, this algorithm is no longer the default
    decomposition algorithm for pydec, as the results it obtains do not make any sense.

    A trivial decomposition algorithm. Just add the output to residual.
    """
    if ref is None:
        recovery = input.c_sum()
    else:
        recovery = ref
    recovery_out = func(recovery)

    if inplace:
        input._component_tensor[:] = 0
        input._residual_tensor = recovery_out
        return input
    else:
        out = pydec.zeros_like(input)
        out._residual_tensor += recovery_out
        return out


@register_decomposition_func("scaling")
def scaling_decomposition(
    input: Composition,
    func: Callable[[Tensor], Tensor],
    *,
    ref: Optional[Tensor] = None,
    inplace: _bool = False,
) -> Composition:
    if ref is None:
        recovery = input.c_sum()
    else:
        recovery = ref
    recovery_out = func(recovery)
    # residual_out = func(input._residual_tensor)
    composition = torch.cat([input.components, input.residual[None]], dim=0)

    multiplier = recovery_out / composition.sum(dim=0)
    multiplier.nan_to_num_(0, 0, 0)

    if inplace:
        input._component_tensor *= multiplier[:-1]
        input._residual_tensor *= multiplier[-1]
        return input
    else:
        out_composition = composition * multiplier
        out_components = out_composition[:-1]
        out_residual = out_composition[-1]
        return pydec.as_composition(out_components, out_residual)



import torch
import pydec
import warnings
import torch.overrides as overrides
import functools
from typing import (
    Any,
    Dict,
    Union,
    List,
    Tuple,
    Sequence,
    Optional,
    Callable,
    TypeVar,
    overload,
)

T = TypeVar("T")

_HANDLED_BUILTIN_FUNCTIONS = {}
_HANDLED_CUSTOMIZED_FUNCTIONS = {}


_namespaces_map = {
    "pydec._composition": torch.Tensor,
    "pydec.variable_functions": torch,
    "pydec.nn": torch.nn,
    "pydec.nn.functional": torch.nn.functional,
}

"""
Functions that are not in the list of overridable functions, but can still
be dispatched via `__torch_function__`.
"""
_whitelist_torch_functions = [
    torch.detach_,
    torch.round_,
    torch.abs_,
    torch.sigmoid_,
    torch.tanh_,
    torch._pack_padded_sequence,
    torch._pad_packed_sequence,
    torch.nn.functional.relu_,
    torch.nn.functional.leaky_relu_,
    torch.reciprocal_,
    torch.exp_,
    torch.sqrt_,
    torch.square_,
]


def _register_function(torch_function: T, builtin: bool = False) -> Callable[[T], T]:
    # checking
    if (
        torch_function not in overrides.get_testing_overrides()
        and torch_function not in _whitelist_torch_functions
    ):
        msg = (
            f"function {torch_function} is not overridable."
            " Invoke `torch.overrides.get_overridable_functions()` to see all overridable torch functions"
        )
        raise TypeError(msg)

    def _decorator(func: Callable):
        if builtin:
            _HANDLED_BUILTIN_FUNCTIONS[torch_function] = func
        else:
            _HANDLED_CUSTOMIZED_FUNCTIONS[torch_function] = func
        return func

    return _decorator


def _register_builtin_function(torch_function: T) -> Callable[[T], T]:
    return _register_function(torch_function, builtin=True)


def register_torch_function(torch_function: T) -> Callable[[T], T]:
    """Register a torch function override for Composition"""
    return _register_function(torch_function, builtin=False)


def _auto_registration(func: T) -> T:
    assert hasattr(func, "__module__")
    assert func.__module__ in _namespaces_map
    torch_namespace = _namespaces_map[func.__module__]
    if not hasattr(torch_namespace, func.__name__):
        msg = (
            f"function ({func.__name__}) is not overridable."
            " Invoke `torch.overrides.get_overridable_functions()` to see all overridable torch functions"
        )
        raise TypeError(msg)
    torch_function = getattr(torch_namespace, func.__name__)

    return _register_function(torch_function, builtin=True)(func)


# def _hadle_self_is_tensor(func: T, inplace=False) -> T:
#     # TODO: inplace warning

#     @functools.wraps(func)
#     def wrapped_func(self, *args, **kwargs):
#         if isinstance(self, torch.Tensor):
#             composition_index = None
#             for i in range(len(args)):
#                 if isinstance(args[i], pydec.Composition):
#                     composition_index = i
#                     break
#             if composition_index is not None:
#                 new_args = list(args)
#                 self, new_args[composition_index] = new_args[composition_index], self
#                 return func(self, *new_args, **kwargs)
#             composition_key = None
#             for k, v in kwargs.items():
#                 if isinstance(v, pydec.Composition):
#                     composition_key = k
#                     break
#             if composition_key is not None:
#                 self, kwargs[composition_key] = kwargs[composition_key], self
#                 return func(self, *args, **kwargs)
#             else:
#                 raise RuntimeError(
#                     f"there is no composition arguments for the invoking of function {func}"
#                 )
#         else:
#             return func(self, *args, **kwargs)

#     return wrapped_func


# def _hadle_self_is_tensor_inplace(func: T) -> T:
#     def wrapped_hadle_self_is_tensor(func: Callable):
#         return _hadle_self_is_tensor(func, inplace=True)

#     return wrapped_hadle_self_is_tensor(func)


def is_registered(torch_function: Callable) -> bool:
    return (
        torch_function in _HANDLED_BUILTIN_FUNCTIONS
        or torch_function in _HANDLED_CUSTOMIZED_FUNCTIONS
    )


def dispatch_torch_function(torch_function: Callable) -> Callable:
    # Customized functions have higher priority
    if torch_function in _HANDLED_CUSTOMIZED_FUNCTIONS:
        return _HANDLED_CUSTOMIZED_FUNCTIONS[torch_function]
    elif torch_function in _HANDLED_BUILTIN_FUNCTIONS:
        return _HANDLED_BUILTIN_FUNCTIONS[torch_function]
    else:
        return None



import torch
from torch._tensor_str import _tensor_str, _add_suffixes, PRINT_OPTS
from typing import List, Tuple, Optional, Union


def _c_add_suffixes(composition_str, suffixes, indent, force_newline):
    tensor_strs = [composition_str]
    last_line_len = len(composition_str) - composition_str.rfind("\n") + 1
    for suffix in suffixes:
        suffix_len = len(suffix)
        if force_newline or last_line_len + suffix_len + 2 > PRINT_OPTS.linewidth:
            tensor_strs.append(",\n" + " " * indent + suffix)
            last_line_len = indent + suffix_len
            force_newline = False
        else:
            tensor_strs.append(", " + suffix)
            last_line_len += suffix_len + 2
    tensor_strs.append("}")
    return "".join(tensor_strs)


def _tensor_str_intern(inp, *, tensor_contents=None, prefix_indent=0):
    is_plain_tensor = type(inp) is torch.Tensor or type(inp) is torch.nn.Parameter
    if inp.is_nested:
        prefix = "nested_tensor("
    elif is_plain_tensor:
        prefix = "tensor("
    else:
        prefix = f"{type(inp).__name__}("

    prefix = " " * prefix_indent + prefix  # add by pydec

    indent = len(prefix)
    suffixes = []
    custom_contents_provided = tensor_contents is not None
    if custom_contents_provided:
        tensor_str = tensor_contents

    # This is used to extract the primal value and thus disable the forward AD
    # within this function.
    # TODO(albanD) This needs to be updated when more than one level is supported
    self, tangent = torch.autograd.forward_ad.unpack_dual(inp)

    # Note [Print tensor device]:
    # A general logic here is we only print device when it doesn't match
    # the device specified in default tensor type.
    # Currently torch.set_default_tensor_type() only supports CPU/CUDA, thus
    # torch._C._get_default_device() only returns either cpu or cuda.
    # In other cases, we don't have a way to set them as default yet,
    # and we should always print out device for them.
    if (
        self.device.type != torch._C._get_default_device()
        or (
            self.device.type == "cuda"
            and torch.cuda.current_device() != self.device.index
        )
        or (self.device.type == "mps")
    ):
        suffixes.append("device='" + str(self.device) + "'")

    # Tensor printing performs tensor operations like slice, indexing, etc to make it in a
    # representable format. These operations on xla/lazy tensor results in compilations. Hence,
    # to avoid compilations, copying the tensor to cpu before printing.
    if self.device.type == "xla" or self.device.type == "lazy":
        self = self.to("cpu")

    # TODO: add an API to map real -> complex dtypes
    _default_complex_dtype = (
        torch.cdouble if torch.get_default_dtype() == torch.double else torch.cfloat
    )
    has_default_dtype = self.dtype in (
        torch.get_default_dtype(),
        _default_complex_dtype,
        torch.int64,
        torch.bool,
    )
    if self.is_sparse:
        suffixes.append("size=" + str(tuple(self.shape)))
        suffixes.append("nnz=" + str(self._nnz()))
        if not has_default_dtype:
            suffixes.append("dtype=" + str(self.dtype))
        if not custom_contents_provided:
            indices_prefix = "indices=tensor("
            indices = self._indices().detach()
            indices_str = _tensor_str(indices, indent + len(indices_prefix))
            if indices.numel() == 0:
                indices_str += ", size=" + str(tuple(indices.shape))
            values_prefix = "values=tensor("
            values = self._values().detach()
            values_str = _tensor_str(values, indent + len(values_prefix))
            if values.numel() == 0:
                values_str += ", size=" + str(tuple(values.shape))
            tensor_str = (
                indices_prefix
                + indices_str
                + "),\n"
                + " " * indent
                + values_prefix
                + values_str
                + ")"
            )
    elif self.layout in {
        torch.sparse_csr,
        torch.sparse_csc,
        torch.sparse_bsr,
        torch.sparse_bsc,
    }:
        suffixes.append("size=" + str(tuple(self.shape)))
        suffixes.append("nnz=" + str(self._nnz()))
        if not has_default_dtype:
            suffixes.append("dtype=" + str(self.dtype))
        if not custom_contents_provided:
            compressed_indices_method, plain_indices_method = {
                torch.sparse_csr: (torch.Tensor.crow_indices, torch.Tensor.col_indices),
                torch.sparse_csc: (torch.Tensor.ccol_indices, torch.Tensor.row_indices),
                torch.sparse_bsr: (torch.Tensor.crow_indices, torch.Tensor.col_indices),
                torch.sparse_bsc: (torch.Tensor.ccol_indices, torch.Tensor.row_indices),
            }[self.layout]
            if self.layout in {torch.sparse_csr, torch.sparse_bsr}:
                cdimname, pdimname = "row", "column"
            else:
                cdimname, pdimname = "column", "row"
            compressed_indices_prefix = f"c{cdimname[:3]}_indices=tensor("
            compressed_indices = compressed_indices_method(self).detach()
            compressed_indices_str = _tensor_str(
                compressed_indices, indent + len(compressed_indices_prefix)
            )
            if compressed_indices.numel() == 0:
                compressed_indices_str += ", size=" + str(
                    tuple(compressed_indices.shape)
                )
            plain_indices_prefix = f"{pdimname[:3]}_indices=tensor("
            plain_indices = plain_indices_method(self).detach()
            plain_indices_str = _tensor_str(
                plain_indices, indent + len(plain_indices_prefix)
            )
            if plain_indices.numel() == 0:
                plain_indices_str += ", size=" + str(tuple(plain_indices.shape))
            values_prefix = "values=tensor("
            values = self.values().detach()
            values_str = _tensor_str(values, indent + len(values_prefix))
            if values.numel() == 0:
                values_str += ", size=" + str(tuple(values.shape))
            tensor_str = (
                compressed_indices_prefix
                + compressed_indices_str
                + "),\n"
                + " " * indent
                + plain_indices_prefix
                + plain_indices_str
                + "),\n"
                + " " * indent
                + values_prefix
                + values_str
                + ")"
            )
    elif self.is_quantized:
        suffixes.append("size=" + str(tuple(self.shape)))
        if not has_default_dtype:
            suffixes.append("dtype=" + str(self.dtype))
        suffixes.append("quantization_scheme=" + str(self.qscheme()))
        if (
            self.qscheme() == torch.per_tensor_affine
            or self.qscheme() == torch.per_tensor_symmetric
        ):
            suffixes.append("scale=" + str(self.q_scale()))
            suffixes.append("zero_point=" + str(self.q_zero_point()))
        elif (
            self.qscheme() == torch.per_channel_affine
            or self.qscheme() == torch.per_channel_symmetric
            or self.qscheme() == torch.per_channel_affine_float_qparams
        ):
            suffixes.append("scale=" + str(self.q_per_channel_scales()))
            suffixes.append("zero_point=" + str(self.q_per_channel_zero_points()))
            suffixes.append("axis=" + str(self.q_per_channel_axis()))
        if not custom_contents_provided:
            tensor_str = _tensor_str(self.dequantize(), indent)
    elif self.is_nested:
        if not custom_contents_provided:

            def indented_str(s, indent):
                return "\n".join(f"  {line}" for line in s.split("\n"))

            strs = ",\n".join(
                indented_str(str(t), indent + 1)
                for t in torch.ops.aten.unbind.int(self, 0)
            )
            tensor_str = f"[\n{strs}\n]"
    else:
        if self.is_meta:
            suffixes.append("size=" + str(tuple(self.shape)))
            if self.dtype != torch.get_default_dtype():
                suffixes.append("dtype=" + str(self.dtype))
            # TODO: This implies that ellipses is valid syntax for allocating
            # a meta tensor, which it could be, but it isn't right now
            if not custom_contents_provided:
                tensor_str = "..."
        else:
            if self.numel() == 0 and not self.is_sparse:
                # Explicitly print the shape if it is not (0,), to match NumPy behavior
                if self.dim() != 1:
                    suffixes.append("size=" + str(tuple(self.shape)))

                # In an empty tensor, there are no elements to infer if the dtype
                # should be int64, so it must be shown explicitly.
                if self.dtype != torch.get_default_dtype():
                    suffixes.append("dtype=" + str(self.dtype))
                if not custom_contents_provided:
                    tensor_str = "[]"
            else:
                if not has_default_dtype:
                    suffixes.append("dtype=" + str(self.dtype))

                if not custom_contents_provided:
                    if self.layout != torch.strided:
                        tensor_str = _tensor_str(self.to_dense(), indent)
                    else:
                        tensor_str = _tensor_str(self, indent)

    if self.layout != torch.strided:
        suffixes.append("layout=" + str(self.layout))

    # Use inp here to get the original grad_fn and not the one generated by the forward grad
    # unpacking.
    if inp.grad_fn is not None:
        name = type(inp.grad_fn).__name__
        if name == "CppFunction":
            name = inp.grad_fn.name().rsplit("::", 1)[-1]
        suffixes.append("grad_fn=<{}>".format(name))
    elif inp.requires_grad:
        suffixes.append("requires_grad=True")

    if self.has_names():
        suffixes.append("names={}".format(self.names))

    if tangent is not None:
        suffixes.append("tangent={}".format(tangent))

    # pydec: ignore the suffixes and return it
    string_repr = _add_suffixes(
        prefix + tensor_str, [], indent, force_newline=self.is_sparse
    )

    # Check if this instance is flagged as a parameter and change the repr accordingly.
    # Unfortunately, this function has to be aware of this detail.
    # NB: This is currently skipped for plain tensor parameters to maintain BC. In the future,
    # this should be done for those as well to produce a valid repr.
    if isinstance(self, torch.nn.Parameter) and not is_plain_tensor:
        # pydec: move the prefix indent to the beginning
        string_repr = " " * prefix_indent + f"Parameter({string_repr[prefix_indent:]})"

    return string_repr, suffixes


def _t_str(
    self: torch.Tensor,
    *,
    tensor_contents: Optional[str] = None,
    prefix_indent: Optional[int] = 0,
) -> Tuple:
    with torch.no_grad():
        return _tensor_str_intern(
            self, tensor_contents=tensor_contents, prefix_indent=prefix_indent
        )


def _c_str(self, *, composition_contents: Optional[List[Union[str, None]]] = None):
    """
    Args:
        self: the composition to be represented
        composition_contents: should be a list of string and the length of the list should be
        equal to the number of components (include the residual). For components that do not
        use customized content, use `None` placeholder in the list.
    """
    if self.components is None:
        return r"composition{}"
    if composition_contents is not None:
        contents_num = len(composition_contents)
        components_num = self.numc() + 1
        if contents_num != components_num:
            if contents_num < components_num:
                composition_contents.extend([None] * (self.numc() + 1 - contents_num))
            else:
                # ignore the redundant
                composition_contents = composition_contents[:components_num]
    prefix = "composition{"
    indent = 2  # fix to 2 looks nice

    components_reprs = [
        _t_str(
            self()[i],
            tensor_contents=composition_contents[i]
            if composition_contents is not None
            else None,
            prefix_indent=indent + 2,  # second indent
        )[0]
        for i in range(self.numc())
    ]
    residual_repr, suffixes = _t_str(
        self._residual_tensor,
        tensor_contents=composition_contents[-1]
        if composition_contents is not None
        else None,
        prefix_indent=indent + 2,  # second indent
    )

    components_str = " " * indent + "components:\n"
    components_str += ",\n".join(components_reprs)

    residual_str = " " * indent + "residual:\n"
    residual_str += residual_repr

    string_repr = _c_add_suffixes(
        prefix + "\n" + components_str + ",\n" + residual_str,
        suffixes,
        indent,
        force_newline=True,
    )

    return string_repr



import types
import pydec
from functools import wraps
from ...autotracing.library import _is_input_composition
from ...exception_utils import args_error_not_in_tracing


class ProxyModule(type):
    r"""This metaclass automatically supports forward computation of
    both tensor and component scopes by replacing the original forward method.
    In addition, it provides a new method to forced convert the
    instances of torch module into instances of pydec module.
    """

    def __new__(cls, clsname, bases, attrs):
        base = bases[0]

        def forward_wrapper(obj, torch_forward, pydec_forward):
            @wraps(torch_forward)
            def meta_forward(self, *args, **kwargs):
                c_mode = _is_input_composition(*args, **kwargs)
                if c_mode and not pydec.is_tracing_enabled():
                    # Maybe just warning.
                    raise args_error_not_in_tracing(*args, **kwargs)
                if c_mode:
                    return pydec_forward(*args, **kwargs)
                else:
                    return torch_forward(*args, **kwargs)

            return types.MethodType(meta_forward, obj)

        user_init = attrs["__init__"] if "__init__" in attrs else None

        def meta_init(self, *args, **kwargs):
            assert isinstance(self, base)
            base.__init__(self, *args, **kwargs)
            if user_init is not None:
                user_init(self, *args, **kwargs)

            torch_forward = self.forward
            pydec_forward = self.pydec_forward
            setattr(
                self, "forward", forward_wrapper(self, torch_forward, pydec_forward)
            )

        def convert_from(cls, obj):
            assert isinstance(obj, base)
            torch_forward = obj.forward
            pydec_forward = types.MethodType(cls.pydec_forward, obj)

            setattr(obj, "forward", forward_wrapper(obj, torch_forward, pydec_forward))
            obj.__class__ = cls
            return obj

        attrs["__init__"] = meta_init

        attrs["convert_from"] = classmethod(convert_from)

        return super().__new__(cls, clsname, bases, attrs)


class CopyModule(type):
    def __new__(cls, clsname, bases, attrs):
        base = bases[0]
        user_init = attrs["__init__"] if "__init__" in attrs else None

        def meta_init(self, *args, **kwargs):
            assert isinstance(self, base)
            base.__init__(self, *args, **kwargs)
            if user_init is not None:
                user_init(self, *args, **kwargs)

        def convert_from(cls, obj):
            assert isinstance(obj, base)
            obj.__class__ = cls
            return obj

        attrs["__init__"] = meta_init

        attrs["convert_from"] = classmethod(convert_from)

        return super().__new__(cls, clsname, bases, attrs)



from __future__ import annotations

import torch
import pydec
import math
import functools

from typing import (
    List,
    Dict,
    Tuple,
    Union,
    Any,
    Callable,
    Optional,
    overload,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from ..._composition import Composition


@functools.lru_cache()
def get_subset(k: int) -> List[List[int]]:
    if k == 0:
        return [[]]
    ssub = get_subset(k - 1)
    sub = []
    for s in ssub:
        sub.append(s.copy())
        sub.append(s + [k - 1])
    return sub


def comb(n, m):
    return math.factorial(n) / math.factorial(m) / math.factorial(n - m)


def c_shapley(
    input: Composition,
    func: Callable,
) -> Composition:
    numc = input.numc()
    out_residual = func(input.residual)
    out_size = func(input.c_sum()).size()
    shapley_values = torch.zeros((numc,) + out_size).to(input)
    subset = get_subset(numc - 1)
    for i in range(numc):
        shapley_i = torch.zeros(out_size).to(input)
        for set in subset:
            indices = torch.LongTensor(set, device=input.device)
            indices[indices >= i] += 1
            mask_input = input.clone()
            mask_input()[indices] = 0
            include_out = func(mask_input.c_sum())
            mask_input()[i] = 0
            exclude_out = func(mask_input.c_sum())
            shapley_i += (include_out - exclude_out) / comb(numc - 1, len(indices))
        shapley_i /= numc
        shapley_values[i] = shapley_i

    return pydec.as_composition(shapley_values, out_residual)



from torch.nn import Module
from typing import TypeVar
from .meta import CopyModule

T = TypeVar("T", bound="DecModule")


class DecModule:
    r"""This base class is mainly used for type hints.
    Most of the processing happens in metaclass `ProxyModule`.
    """

    def pydec_forward(self, *args, **kwargs):
        r"""Defines the computation performed in tracing.

        Should be overridden by all subclasses.
        """
        raise NotImplementedError(
            f'Module [{type(self).__name__}] is missing the required "pydec_forward" function'
        )

    def convert_from(cls: T, obj) -> T:
        r"""Converting a torch Module to pydec Module.

        Warning: The `__init__` of the pydec Module will not be invoked.
        Just use it in auto tracing.
        """
        raise NotImplementedError(
            f'Module [{cls.__name__}] is missing the required "convert_from" function'
        )


class Module(Module, DecModule, metaclass=CopyModule):
    __doc__ = Module.__doc__



import pydec
from pydec import Composition
import torch
import pytest
from ._composition_test import init_composition

torch.manual_seed(114514)


class TestCat:
    c_list = [init_composition((i + 1, 2)) for i in range(3)]

    def test1(self):
        out = pydec.cat(self.c_list, dim=0)
        assert out.c_size() == (3, 6, 2)

    def test2(self):
        out_holder = init_composition((6, 2))
        out = pydec.cat(self.c_list, dim=0, out=out_holder)
        assert out_holder.c_size() == (3, 6, 2)
        assert (out == out_holder).all()

    def test3(self):
        with pytest.raises(RuntimeError):
            pydec.cat(tuple(self.c_list), dim=1)


class TestCCat:
    c_list = [init_composition((3, 2), i + 1) for i in range(3)]

    def test1(self):
        out = pydec.c_cat(self.c_list)
        assert out.c_size() == (6, 3, 2)

    def test2(self):
        out_holder = init_composition((3, 2), 6)
        out = pydec.c_cat(self.c_list, out=out_holder)
        assert out_holder.c_size() == (6, 3, 2)
        assert (out._component_tensor == out_holder._component_tensor).all()

    def test3(self):
        with pytest.raises(RuntimeError):
            pydec.c_cat(self.c_list + [init_composition((2, 2))])


class TestStack:
    c_list = [init_composition((2, 2)) for i in range(3)]

    def test1(self):
        out = torch.stack(self.c_list, dim=0)
        assert out.c_size() == (3, 3, 2, 2)

    def test2(self):
        out_holder = init_composition((3, 2, 2))
        out = torch.stack(self.c_list, dim=0, out=out_holder)
        assert out_holder.c_size() == (3, 3, 2, 2)
        assert (out == out_holder).all()

    def test3(self):
        with pytest.raises(RuntimeError):
            torch.stack(self.c_list + [init_composition((3, 2))], dim=0)


class TestCStack:
    t_list = [torch.rand((2, 2)) for i in range(3)]

    def test1(self):
        out = pydec.c_stack(self.t_list)
        assert out.c_size() == (3, 2, 2)

    def test2(self):
        out_holder = pydec.zeros((2, 2), 3)
        out = pydec.c_stack(self.t_list, out=out_holder)
        assert out_holder.c_size() == (3, 2, 2)
        assert (out._component_tensor == out_holder._component_tensor).all()

    def test3(self):
        with pytest.raises(RuntimeError):
            pydec.c_cat(self.t_list + [torch.rand((3, 2))])


class TestDiagonalInit:
    c = init_composition((2, 3))

    def test1(self):
        data = torch.zeros((2, 3))

        out = pydec.diagonal_init(self.c, data, 1)

        for i in range(self.c.numc()):
            for j in range(self.c.numc()):
                if i == j:
                    assert torch.all(out()[i, :, j] == data[:, i])
                else:
                    assert torch.all(out()[i, :, j] == self.c()[i, :, j])

    def test2(self):
        c = init_composition((2, 3), 4)

        data = torch.zeros((2, 3))
        offset = -1

        out = pydec.diagonal_init(c, data, 1, offset=offset)

        for i in range(1, c.numc()):
            for j in range(c.size(1)):
                if i + offset == j:
                    assert torch.all(out()[i, :, j] == data[:, j])
                else:
                    assert torch.all(out()[i, :, j] == c()[i, :, j])

    def test3(self):
        component_num = 3
        size = (3, 2)
        x = torch.randn(size)
        c = pydec.zeros((3, 2), component_num)
        out = pydec.diagonal_init(c, src=x, dim=0)
        for i in range(self.c.numc()):
            for j in range(self.c.numc()):
                if i == j:
                    assert torch.all(out()[i, j] == x[i])
                else:
                    assert torch.all(out()[i, j] == c()[i, j])


class TestCApply:
    def test_abs(self):
        c = init_composition((3, 4))
        out = pydec.c_apply(c, torch.abs)
        assert (out._component_tensor == c._component_tensor.abs()).all()


class TestCMap:
    def test_add(self):
        c1 = init_composition((3, 4))
        c2 = init_composition((4,))
        out = pydec.c_map(c1, c2, torch.add)
        assert (
            out._component_tensor == c1._component_tensor + c2._component_tensor
        ).all()


class TestNumel:
    def test1(self):
        c1 = init_composition((3, 4))
        assert c1.numel() == 3 * 4


class TestCNumel:
    def test1(self):
        c1 = init_composition((3, 4))
        assert c1.c_numel() == 3 * 4 * c1.numc()


class TestNumc:
    def test1(self):
        c1 = init_composition((3, 4), 10)
        assert c1.numc() == 10


class TestClone:
    def test1(self):
        c1 = init_composition((3, 4), 10)
        c1[:] = 1.1
        c1.residual[:] = 3.5
        c2 = torch.clone(c1)
        assert (c1.components == c2.components).all()
        assert (c1.residual == c2.residual).all()
        c1[:] = 2.0
        c1.residual[:] = 4.0
        assert not (c1.components == c2.components).any()
        assert not (c1.residual == c2.residual).any()


class TestDetach:
    def test1(self):
        size = (3, 4)
        numc = 10
        c1 = init_composition(size, numc)
        leaf_tensors = []
        for i in range(numc):
            leaf_tensors.append(torch.randn(size, requires_grad=True))
            c1()[i] = leaf_tensors[i]
        loss = (c1 * 2 + torch.detach(c1 * 3)).c_sum().sum()
        loss.backward()
        print(leaf_tensors[0].grad)
        assert (leaf_tensors[0].grad == 2).all()

    def test2(self):
        size = (3, 4)
        numc = 10
        c1 = init_composition(size, numc)
        leaf_tensors = []
        for i in range(numc):
            leaf_tensors.append(torch.randn(size, requires_grad=True))
            c1()[i] = leaf_tensors[i]
        loss = (c1 * 2 + torch.detach_(c1 * 3)).c_sum().sum()
        loss.backward()
        print(leaf_tensors[0].grad)
        assert (leaf_tensors[0].grad == 2).all()


class TestAdd:
    def test1(self):
        c1 = init_composition((3, 4))
        c2 = init_composition((4,))
        out = torch.add(c1, c2)
        assert (out.components == c1.components + c2.components).all()

    def test2(self):
        c1 = init_composition((3, 4))
        c2 = init_composition((4,))
        out = torch.add(c1, c2, alpha=2)
        assert (out.components == c1.components + 2 * c2.components).all()

    def test3(self):
        c1 = init_composition((3, 4))
        t1 = torch.rand((4,))
        out = torch.add(c1, t1)
        assert (out.components == c1.components).all()
        assert (out.residual == t1).all()

    def test4(self):
        c1 = init_composition((3, 4))
        c2 = init_composition((4,))
        out_holder = pydec.zeros((3, 4), 3)
        out = torch.add(c1, c2, out=out_holder)
        assert (out._component_tensor == out_holder._component_tensor).all()


class TestVar:
    def test1(self):
        c_input = init_composition((2, 3, 4))
        input = c_input.c_sum()

        ref = torch.var(input)
        c_out = torch.var(c_input)
        assert (ref - c_out.c_sum()).abs().sum() < 1e-3

        ref = torch.var(input, False)
        c_out = torch.var(c_input, False)
        assert (ref - c_out.c_sum()).abs().sum() < 1e-3

    def test2(self):
        c_input = init_composition((2, 3, 4))
        input = c_input.c_sum()
        ref = torch.var(input, -1)
        c_out = torch.var(c_input, -1)
        assert (ref - c_out.c_sum()).abs().sum() < 1e-3

        ref = torch.var(input, -1, unbiased=False)
        c_out = torch.var(c_input, -1, unbiased=False)
        assert (ref - c_out.c_sum()).abs().sum() < 1e-3

    def test3(self):
        c_input = init_composition((2, 3, 4))
        input = c_input.c_sum()
        ref = torch.var(input, dim=(0, 2))
        c_out = torch.var(c_input, dim=(0, 2))
        assert (ref - c_out.c_sum()).abs().sum() < 1e-3

        ref = torch.var(input, dim=(0, 2), unbiased=False)
        c_out = torch.var(c_input, dim=(0, 2), unbiased=False)
        assert (ref - c_out.c_sum()).abs().sum() < 1e-3


class TestBmm:
    def test1(self):
        c_input1 = init_composition((2, 3, 4))
        c_input2 = init_composition((2, 4, 5))
        input1 = c_input1.c_sum()
        input2 = c_input2.c_sum()
        ref = torch.bmm(input1, input2)
        c_out = torch.bmm(c_input1, c_input2)
        assert (ref - c_out.c_sum()).abs().sum() < 1e-3



from .meta import ProxyModule, CopyModule
from .module import DecModule
from pydec import Composition
import torch.nn as nn
from .. import functional as F
from typing import (
    List,
    Tuple,
    Optional,
    Union,
    Any,
    ContextManager,
    Callable,
    overload,
    Iterator,
    NamedTuple,
    Sequence,
    TypeVar,
)

__all__ = [
    "Identity",
    "Linear",
]


class Identity(nn.Identity, DecModule, metaclass=CopyModule):
    __doc__ = nn.Identity.__doc__


class Linear(nn.Linear, DecModule, metaclass=ProxyModule):
    def pydec_forward(self, input: Composition) -> Composition:
        return F.linear(input, weight=self.weight, bias=self.bias)



import torch
import pydec
from torch import Tensor
from torch.nn.parameter import Parameter
from torch._C import memory_format

import pydec.core as core
from pydec._composition import Composition, IndexComposition
from .overrides import _auto_registration, _register_builtin_function

from typing import (
    Any,
    Union,
    List,
    Tuple,
    Optional,
    Callable,
    overload,
)

# In some cases, these basic types are shadowed by corresponding
# top-level values.  The underscore variants let us refer to these
# types.  See https://github.com/python/mypy/issues/4146 for why these
# workarounds is necessary
# In the future, this will become useful if mypy is introduced into pydec
from torch.types import (
    _int,
    _float,
    _bool,
    Number,
    _dtype,
    _device,
    _qscheme,
    _size,
    _layout,
)

from torch import strided

from pydec.utils import _shift_dim, _shift_dims
from pydec.exception_utils import (
    arg_value_error,
    none_decomposition_func_error,
    unsupported_operand_error,
)
import builtins


def void() -> Composition:
    return Composition()


@overload
def as_composition(components: Tensor, residual: Tensor = None) -> Composition:
    ...


def as_composition(
    components: Tensor, residual: Tensor = None, check: _bool = True
) -> Composition:
    # TODO: implement `check`, check shape, dtype and device
    out = void()
    out._component_tensor = components
    if residual is None:
        residual = torch.zeros(components.size()[1:]).to(components)
    out._residual_tensor = residual
    return out


@_auto_registration
def cat(
    compositions: Union[Tuple[Composition, ...], List[Composition]],
    dim: _int = 0,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    c_tensors = tuple(c._component_tensor for c in compositions)
    out_component_tensor = torch.cat(
        c_tensors,
        _shift_dim(dim),
        out=out._component_tensor if out is not None else None,
    )
    r_tensors = tuple(c._residual_tensor for c in compositions)
    out_residual_tensor = torch.cat(
        r_tensors,
        dim,
        out=out._residual_tensor if out is not None else None,
    )
    return as_composition(out_component_tensor, out_residual_tensor)


@_auto_registration
def concat(
    compositions: Union[Tuple[Composition, ...], List[Composition]],
    dim: _int = 0,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    return cat(compositions=compositions, dim=dim, out=out)


def concatenate(
    compositions: Union[Tuple[Composition, ...], List[Composition]],
    dim: _int = 0,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    return cat(compositions=compositions, dim=dim, out=out)


def c_cat(
    compositions: Union[Tuple[Composition, ...], List[Composition]],
    *,
    sum_residual: _bool = True,
    out: Optional[Composition] = None,
) -> Composition:
    for i in range(1, len(compositions)):
        if compositions[i].size() != compositions[0].size():
            raise arg_value_error(
                f"Sizes of compositions must match except the number of composition. Expected size [{compositions[0].size()}] but got size [{compositions[i].size()}] for component number {i} in the list"
            )

    c_tensors = tuple(c._component_tensor for c in compositions)
    out_component_tensor = torch.cat(
        c_tensors,
        0,
        out=out._component_tensor if out is not None else None,
    )
    out_residual_tensor = None
    if sum_residual:
        r_tensors = tuple(c._residual_tensor for c in compositions)
        out_residual_tensor = builtins.sum(r_tensors)
    return as_composition(out_component_tensor, out_residual_tensor)


@_auto_registration
def stack(
    compositions: Union[Tuple[Composition, ...], List[Composition]],
    dim: _int = 0,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    for i in range(1, len(compositions)):
        if compositions[i].size() != compositions[0].size():
            raise arg_value_error(
                f"Sizes of compositions must match except the number of composition. Expected size [{compositions[0].size()}] but got size [{compositions[i].size()}] for component number {i} in the list"
            )
    c_tensors = tuple(c._component_tensor for c in compositions)
    out_component_tensor = torch.stack(
        c_tensors,
        _shift_dim(dim),
        out=out._component_tensor if out is not None else None,
    )
    r_tensors = tuple(c._residual_tensor for c in compositions)
    out_residual_tensor = torch.stack(
        r_tensors,
        dim,
        out=out._residual_tensor if out is not None else None,
    )
    return as_composition(out_component_tensor, out_residual_tensor)


def c_stack(
    components: Union[Tuple[Tensor, ...], List[Tensor]],
    *,
    out: Optional[Composition] = None,
) -> Composition:
    for i in range(1, len(components)):
        if components[i].size() != components[0].size():
            raise arg_value_error(
                f"Sizes of components must match. Expected size [{components[0].size()}] but got size [{components[i].size()}] for component number {i} in the list"
            )

    out_component_tensor = torch.stack(
        components,
        0,
        out=out._component_tensor if out is not None else None,
    )
    return as_composition(out_component_tensor)


def diagonal_init(
    input: Composition, src: Tensor, dim: _int, offset: _int = 0
) -> Composition:
    permute_dims = list(range(src.dim()))
    dim = (dim + src.dim()) % src.dim()  # Converted to a positive number
    permute_dims.remove(dim)
    permute_dims.append(dim)
    src = src.permute(permute_dims)
    if (
        torch.__version__ < "1.11.0"
    ):  # for versions < 1.11.0, 'diagonal_scatter' does not exist.
        out_component_tensor = input._component_tensor.clone()
        diag_view = out_component_tensor.diagonal(
            offset=offset, dim1=0, dim2=_shift_dim(dim)
        )
        diag_view[:] = src
    else:
        out_component_tensor = input._component_tensor.diagonal_scatter(
            src, offset=offset, dim1=0, dim2=_shift_dim(dim)
        )
    out_residual_tensor = input._residual_tensor.clone()

    return as_composition(out_component_tensor, out_residual_tensor)


def c_apply(input: Composition, callable: Callable[..., Tensor]) -> Composition:
    # TODO: check shape, dtype and device of result
    out_component_tensor = callable(input._component_tensor)
    out_residual_tensor = callable(input._residual_tensor)
    if not isinstance(out_component_tensor, Tensor) or not isinstance(
        out_residual_tensor, Tensor
    ):
        raise RuntimeError(
            "TypeError: 'callable' must return a tensor, not {}".format(
                type(
                    out_component_tensor
                    if not isinstance(out_component_tensor, Tensor)
                    else out_residual_tensor
                ).__name__
            )
        )
    return as_composition(out_component_tensor, out_residual_tensor)


def c_map(input, other: Composition, callable: Callable[..., Tensor]) -> Composition:
    # TODO: check shape, dtype and device of result
    out_component_tensor = callable(input._component_tensor, other._component_tensor)
    out_residual_tensor = callable(input._residual_tensor, other._residual_tensor)
    if not isinstance(out_component_tensor, Tensor) or not isinstance(
        out_residual_tensor, Tensor
    ):
        raise RuntimeError(
            "TypeError: 'callable' must return a tensor, not {}".format(
                type(
                    out_component_tensor
                    if not isinstance(out_component_tensor, Tensor)
                    else out_residual_tensor
                ).__name__
            )
        )
    return as_composition(out_component_tensor, out_residual_tensor)


@_auto_registration
def numel(input: Composition) -> _int:
    return torch.numel(input._residual_tensor)


def c_numel(input: Composition, count_residual=False) -> _int:
    if count_residual:
        return input._component_tensor.numel() + input._residual_tensor.numel()
    else:
        return input._component_tensor.numel()


def numc(input: Composition) -> _int:
    return len(input.components)


@_auto_registration
def clone(
    input: Composition,
    *,
    memory_format: Optional[memory_format] = None,
) -> Composition:
    out_component_tensor = torch.clone(
        input._component_tensor, memory_format=memory_format
    )
    out_residual_tensor = torch.clone(
        input._residual_tensor, memory_format=memory_format
    )
    return as_composition(out_component_tensor, out_residual_tensor)


@_auto_registration
def detach(input: Composition) -> Composition:
    out_component_tensor = torch.detach(input._component_tensor)
    out_residual_tensor = torch.detach(input._residual_tensor)
    return as_composition(out_component_tensor, out_residual_tensor)


@_auto_registration
def detach_(input: Composition) -> Composition:
    torch.detach_(input._component_tensor)
    torch.detach_(input._residual_tensor)
    return input


@overload
def add(
    input: Composition,
    other: Union[Composition, Tensor, Number],
    *,
    alpha: Optional[Number] = 1,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def add(
    input: Union[Composition, Tensor, Number],
    other: Composition,
    *,
    alpha: Optional[Number] = 1,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@_auto_registration
def add(
    input: Composition,
    other: Union[Composition, Tensor, Number],
    *,
    alpha: Optional[Number] = 1,
    out: Optional[Composition] = None,
) -> Composition:
    # TODO: fix bug: broadcasting errors, trigger when composition + composition
    # TODO: fix bug: shape errors, trigger when composition(residual broadcasting) + tensor
    # TODO: fix bug: also fix `sub`
    # TODO: add broadcasting examples in docs after bug fixing
    if not isinstance(input, Composition) and not isinstance(other, Composition):
        raise unsupported_operand_error("add", type(input), type(other))
    if isinstance(input, Composition):
        if isinstance(other, Composition):
            return core.decBLAS.cc_add(input, other, alpha=alpha, out=out)
        elif isinstance(other, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_add(input, other, alpha=alpha, out=out)
        else:
            raise unsupported_operand_error("add", type(input), type(other))
    elif isinstance(input, (_int, _float, _bool, Tensor)):
        return core.decBLAS.tc_add(input, other, alpha=alpha, out=out)
    else:
        raise unsupported_operand_error("add", type(input), type(other))


@overload
def sub(
    input: Composition,
    other: Union[Composition, Tensor, Number],
    *,
    alpha: Optional[Number] = 1,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def sub(
    input: Union[Composition, Tensor, Number],
    other: Composition,
    *,
    alpha: Optional[Number] = 1,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@_auto_registration
def sub(
    input: Union[Composition, Tensor, Number],
    other: Union[Composition, Tensor, Number],
    *,
    alpha: Optional[Number] = 1,
    out: Optional[Composition] = None,
) -> Composition:
    # TODO: fix bug: see TODOs in `add`
    if not isinstance(input, Composition) and not isinstance(other, Composition):
        raise unsupported_operand_error("sub", type(input), type(other))
    if isinstance(input, Composition):
        if isinstance(other, Composition):
            return core.decBLAS.cc_sub(input, other, alpha=alpha, out=out)
        elif isinstance(other, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_sub(input, other, alpha=alpha, out=out)
        else:
            raise unsupported_operand_error("sub", type(input), type(other))
    elif isinstance(input, (_int, _float, _bool, Tensor)):
        return core.decBLAS.tc_sub(input, other, alpha=alpha, out=out)
    else:
        raise unsupported_operand_error("sub", type(input), type(other))


@overload
def subtract(
    input: Composition,
    other: Tensor,
    *,
    alpha: Number = 1,
    out: Optional[Tensor] = None,
) -> Composition:
    ...


@overload
def subtract(input: Composition, other: Number, alpha: Number = 1) -> Composition:
    ...


@_auto_registration
def subtract(*args, **kwargs) -> Composition:
    return sub(*args, **kwargs)


@overload
def mul(
    input: Composition,
    other: Union[Tensor, Number],
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def mul(
    input: Union[Tensor, Number],
    other: Composition,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def mul(
    input: Composition,
    other: Composition,
    *,
    out: Optional[Composition] = None,
    ref_input: Optional[Tensor] = None,
    ref_other: Optional[Tensor] = None,
) -> Composition:
    ...


@_auto_registration
def mul(
    input: Union[Composition, Tensor, Number],
    other: Union[Composition, Tensor, Number],
    *,
    out: Optional[Composition] = None,
    ref_input: Optional[Tensor] = None,
    ref_other: Optional[Tensor] = None,
) -> Composition:
    if not isinstance(input, Composition) and not isinstance(other, Composition):
        raise unsupported_operand_error("mul", type(input), type(other))
    if isinstance(input, Composition):
        if isinstance(other, Composition):
            return core.decMVF.cc_mul(
                input,
                other,
                out=out,
                ref_input=ref_input,
                ref_other=ref_other,
            )
        elif isinstance(other, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_mul(input, other, out=out)
        else:
            raise unsupported_operand_error("mul", type(input), type(other))
    elif isinstance(input, (_int, _float, _bool, Tensor)):
        return core.decBLAS.tc_mul(input, other, out=out)
    else:
        raise unsupported_operand_error("mul", type(input), type(other))


@overload
def multiply(
    input: Composition, other: Tensor, *, out: Optional[Tensor] = None
) -> Composition:
    ...


@overload
def multiply(input: Composition, other: Number) -> Composition:
    ...


@_auto_registration
def multiply(
    input: Composition, other: Any, *, out: Optional[Tensor] = None
) -> Composition:
    return mul(input, other=other, out=out)


@overload
def div(
    input: Composition,
    other: Union[Tensor, Number],
    *,
    rounding_mode: Optional[str] = None,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def div(
    input: Union[Tensor, Number],
    other: Composition,
    *,
    rounding_mode: Optional[str] = None,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def div(
    input: Composition,
    other: Composition,
    *,
    rounding_mode: Optional[str] = None,
    out: Optional[Composition] = None,
    ref_input: Optional[Tensor] = None,
    ref_other: Optional[Tensor] = None,
) -> Composition:
    ...


@_auto_registration
def div(
    input,
    other,
    *,
    rounding_mode: Optional[str] = None,
    out: Optional[Composition] = None,
    ref_input: Optional[Tensor] = None,
    ref_other: Optional[Tensor] = None,
) -> Composition:
    if not isinstance(input, Composition) and not isinstance(other, Composition):
        raise unsupported_operand_error("div", type(input), type(other))
    if isinstance(input, Composition):
        if isinstance(other, Composition):
            return core.decMVF.cc_div(
                input,
                other,
                rounding_mode=rounding_mode,
                ref_input=ref_input,
                ref_other=ref_other,
            )
        elif isinstance(other, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_div(
                input, other, rounding_mode=rounding_mode, out=out
            )
        else:
            raise unsupported_operand_error("div", type(input), type(other))
    elif isinstance(input, (_int, _float, _bool, Tensor)):
        # TODO: not implement yet
        raise unsupported_operand_error("div", type(input), type(other))
    else:
        raise unsupported_operand_error("div", type(input), type(other))


@overload
def divide(
    input: Composition,
    other: Tensor,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def divide(
    input: Composition,
    other: Tensor,
    *,
    rounding_mode: Optional[str],
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def divide(
    input: Composition, other: Number, *, rounding_mode: Optional[str]
) -> Composition:
    ...


@overload
def divide(
    input: Composition,
    other: Number,
) -> Composition:
    ...


@_auto_registration
def divide(*args, **kwargs):
    return div(*args, **kwargs)


@overload
def mv(
    input: Composition,
    vec: Union[Composition, Tensor],
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def mv(
    input: Union[Composition, Tensor],
    vec: Composition,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@_auto_registration
def mv(
    input,
    vec,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    if not isinstance(input, Composition) and not isinstance(vec, Composition):
        raise unsupported_operand_error("mv", type(input), type(vec))
    if isinstance(input, Composition):
        if isinstance(vec, Composition):
            # TODO: not implement yet
            raise unsupported_operand_error("mv", type(input), type(vec))
        elif isinstance(vec, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_mv(input, vec, out=out)
        else:
            raise unsupported_operand_error("mv", type(input), type(vec))
    elif isinstance(input, (_int, _float, _bool, Tensor)):
        return core.decBLAS.ct_mv(input, vec, out=out)
    else:
        raise unsupported_operand_error("mv", type(input), type(vec))


@overload
def mm(
    input: Composition,
    mat2: Union[Composition, Tensor],
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def mm(
    input: Union[Composition, Tensor],
    mat2: Composition,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@_auto_registration
def mm(
    input,
    mat2,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    if not isinstance(input, Composition) and not isinstance(mat2, Composition):
        raise unsupported_operand_error("mm", type(input), type(mat2))
    if isinstance(input, Composition):
        if isinstance(mat2, Composition):
            # TODO: not implement yet
            raise unsupported_operand_error("mm", type(input), type(mat2))
        elif isinstance(mat2, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_mm(input, mat2, out=out)
        else:
            raise unsupported_operand_error("mm", type(input), type(mat2))
    elif isinstance(input, (_int, _float, _bool, Tensor)):
        return core.decBLAS.tc_mm(input, mat2, out=out)
    else:
        raise unsupported_operand_error("mm", type(input), type(mat2))


@overload
def bmm(
    input: Composition,
    mat2: Union[Composition, Tensor],
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def bmm(
    input: Union[Composition, Tensor],
    mat2: Composition,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@_auto_registration
def bmm(
    input: Union[Composition, Tensor],
    mat2: Union[Composition, Tensor],
    *,
    out: Optional[Composition] = None,
) -> Composition:
    if not isinstance(input, Composition) and not isinstance(mat2, Composition):
        raise unsupported_operand_error("bmm", type(input), type(mat2))
    if isinstance(input, Composition):
        if isinstance(mat2, Composition):
            return core.decMVF.cc_bmm(input, mat2, out=out)
        elif isinstance(mat2, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_bmm(input, mat2, out=out)
        else:
            raise unsupported_operand_error("bmm", type(input), type(mat2))
    elif isinstance(input, Tensor):
        assert isinstance(mat2, Composition)  # narrow type
        return core.decBLAS.tc_bmm(input, mat2, out=out)
    else:
        raise unsupported_operand_error("bmm", type(input), type(mat2))


@overload
def matmul(
    input: Composition,
    other: Union[Composition, Tensor],
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def matmul(
    input: Union[Composition, Tensor],
    other: Composition,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@_auto_registration
def matmul(
    input: Union[Composition, Tensor],
    other: Union[Composition, Tensor],
    *,
    out: Optional[Composition] = None,
) -> Composition:
    if not isinstance(input, Composition) and not isinstance(other, Composition):
        raise unsupported_operand_error("matmul", type(input), type(other))
    if isinstance(input, Composition):
        if isinstance(other, Composition):
            return core.decMVF.cc_matmul(input, other)
        elif isinstance(other, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_matmul(input, other, out=out)
        else:
            raise unsupported_operand_error("matmul", type(input), type(other))
    elif isinstance(input, (_int, _float, _bool, Tensor)):
        return core.decBLAS.tc_matmul(input, other, out=out)
    else:
        raise unsupported_operand_error("matmul", type(input), type(other))


@overload
def any(input: Composition, *, out: Optional[Tensor] = None) -> Tensor:
    ...


@overload
def any(
    input: Composition,
    dim: _int,
    keepdim: _bool = False,
    *,
    out: Optional[Tensor] = None,
) -> Tensor:
    ...


@_auto_registration
def any(input: Composition, *args: Any, **kwargs: Any) -> Tensor:
    return torch.any(input.c_sum(), *args, **kwargs)


@overload
def all(input: Composition, *, out: Optional[Tensor] = None) -> Tensor:
    ...


@overload
def all(
    input: Composition,
    dim: _int,
    keepdim: _bool = False,
    *,
    out: Optional[Tensor] = None,
) -> Tensor:
    ...


@_auto_registration
def all(input: Composition, *args: Any, **kwargs: Any) -> Tensor:
    return torch.all(input.c_sum(), *args, **kwargs)


@_auto_registration
def isinf(input: Composition) -> Composition:
    out_residual = torch.isinf(input.residual)
    out_components = torch.isinf(input.components)
    return as_composition(out_components, out_residual)


@_auto_registration
def isnan(input: Composition) -> Composition:
    out_residual = torch.isnan(input.residual)
    out_components = torch.isnan(input.components)
    return as_composition(out_components, out_residual)


@_auto_registration
def unsqueeze(input: Composition, dim: _int) -> Composition:
    out_residual_tensor = input._residual_tensor.unsqueeze(dim)
    out_component_tensor = input._component_tensor.unsqueeze(_shift_dim(dim))
    return as_composition(out_component_tensor, out_residual_tensor)


@overload
def squeeze(input: Composition) -> Composition:
    ...


@overload
def squeeze(input: Composition, dim: _int) -> Composition:
    ...


@_auto_registration
def squeeze(input: Composition, dim=None) -> Composition:
    if dim is None:
        out_residual_tensor = input._residual_tensor.squeeze()
        out_component_tensor = input._component_tensor.squeeze()
        if input.numc() == 1:
            out_component_tensor = out_component_tensor.unsqueeze(0)
    else:
        out_residual_tensor = input._residual_tensor.squeeze(dim)
        out_component_tensor = input._component_tensor.squeeze(_shift_dim(dim))
    return as_composition(out_component_tensor, out_residual_tensor)


@_auto_registration
def transpose(input: Composition, dim0: _int, dim1: _int) -> Composition:
    out_residual_tensor = input._residual_tensor.transpose(dim0, dim1)
    out_component_tensor = input._component_tensor.transpose(
        _shift_dim(dim0), _shift_dim(dim1)
    )
    return as_composition(out_component_tensor, out_residual_tensor)


@_auto_registration
def permute(input: Composition, dims: _size) -> Composition:
    out_residual_tensor = input._residual_tensor.permute(dims)
    out_component_tensor = input._component_tensor.permute((0,) + _shift_dims(dims))
    return as_composition(out_component_tensor, out_residual_tensor)


@overload
def sum(input: Composition, *, dtype: Optional[_dtype] = None) -> Composition:
    ...


@overload
def sum(
    input: Composition,
    dim: Union[_int, _size],
    keepdim: _bool = False,
    *,
    dtype: Optional[_dtype] = None,
    out: Optional[Composition] = None,
) -> Composition:
    ...


# TODO: to support 'out: Optional[Composition] = None'
@_auto_registration
def sum(
    input: Composition,
    dim=None,
    keepdim: _bool = False,
    *,
    dtype: Optional[_dtype] = None,
    out: Optional[Composition] = None,
) -> Composition:
    out_components = {}
    out_residual = {}
    if out is not None:
        out_components["out"] = out.components
        out_residual["out"] = out.residual
    if dim is None:
        if input.ndim == 0:
            return input
        dim = tuple(range(1, input._component_tensor.dim()))
        out_component_tensor = torch.sum(
            input._component_tensor,
            dim=dim,
            dtype=dtype,
            **out_components,
        )
        out_residual_tensor = torch.sum(
            input._residual_tensor, dtype=dtype, **out_residual
        )
    else:
        out_residual_tensor = torch.sum(
            input._residual_tensor,
            dim=dim,
            keepdim=keepdim,
            dtype=dtype,
            **out_residual,
        )
        if isinstance(dim, _int):
            dim = (dim,)
        out_component_tensor = torch.sum(
            input._component_tensor,
            dim=_shift_dims(dim),
            keepdim=keepdim,
            dtype=dtype,
            **out_components,
        )
    return as_composition(out_component_tensor, out_residual_tensor)


def c_sum(input: Composition, *, dtype: Optional[_dtype] = None) -> Tensor:
    return input._component_tensor.sum(dim=0, dtype=dtype) + input._residual_tensor.to(
        dtype=dtype
    )


@overload
def mean(input: Composition, *, dtype: Optional[_dtype] = None) -> Composition:
    ...


@overload
def mean(
    input: Composition,
    dim: Union[_int, _size],
    keepdim: _bool = False,
    *,
    dtype: Optional[_dtype] = None,
    out: Optional[Composition] = None,
) -> Composition:
    ...


# TODO: to support 'out: Optional[Composition] = None'
@_auto_registration
def mean(
    input: Composition,
    dim: Union[None, _int, _size] = None,
    keepdim: _bool = False,
    *,
    dtype: Optional[_dtype] = None,
    out: Optional[Composition] = None,
) -> Composition:
    # TODO: move to BLAS
    if out is not None:
        raise arg_value_error(
            f"{mean.__name__}() dees not support keyword 'out' currently"
        )
    if dim is None:
        dim = tuple(range(1, input._component_tensor.dim()))
        if out is not None:
            out_component_tensor = torch.mean(
                input._component_tensor, dim=dim, dtype=dtype, out=out._component_tensor
            )
            out_residual_tensor = torch.mean(
                input._residual_tensor, dtype=dtype, out=out._residual_tensor
            )
        else:
            out_component_tensor = torch.mean(
                input._component_tensor, dim=dim, dtype=dtype
            )
            out_residual_tensor = torch.mean(input._residual_tensor, dtype=dtype)
    else:
        out_residual_tensor = torch.mean(
            input._residual_tensor,
            dim=dim,
            keepdim=keepdim,
            dtype=dtype,
            out=out._residual_tensor if out is not None else None,
        )
        if isinstance(dim, _int):
            dim = (dim,)
        out_component_tensor = torch.mean(
            input._component_tensor,
            dim=_shift_dims(dim),
            keepdim=keepdim,
            dtype=dtype,
            out=out._component_tensor if out is not None else None,
        )
    return as_composition(out_component_tensor, out_residual_tensor)


@_auto_registration
def reshape(input: Composition, shape: _size) -> Composition:
    out_component_tensor = input._component_tensor.reshape((input.numc(),) + shape)
    out_residual_tensor = input._residual_tensor.reshape(shape)
    return as_composition(out_component_tensor, out_residual_tensor)


@overload
def masked_fill(input: Composition, mask: Tensor, value: Tensor) -> Composition:
    ...


@overload
def masked_fill(input: Composition, mask: Tensor, value: Number) -> Composition:
    ...


@_auto_registration
def masked_fill(input: Composition, mask: Tensor, value: Any) -> Composition:
    out_component_tensor = input._component_tensor.masked_fill(mask[None], value)
    out_residual_tensor = input._residual_tensor.masked_fill(mask, value)
    return as_composition(out_component_tensor, out_residual_tensor)


@overload
def c_masked_fill(input: Composition, mask: Tensor, value: Tensor) -> Composition:
    ...


@overload
def c_masked_fill(input: Composition, mask: Tensor, value: Number) -> Composition:
    ...


def c_masked_fill(input: Composition, mask: Tensor, value: Any) -> Composition:
    if mask.dim() == 1:
        if len(mask) != input.numc():
            raise arg_value_error(
                f"the length of mask ({len(mask)}) should match component number ({input.numc()})"
            )
        mask_size = (input.numc(),) + (1,) * input.dim()
        mask = mask.view(mask_size)
    out_component_tensor = input._component_tensor.masked_fill(mask, value)
    out_residual_tensor = input._residual_tensor.clone()
    return as_composition(out_component_tensor, out_residual_tensor)


# TODO: to support 'out: Optional[Composition] = None'
@_auto_registration
def masked_select(
    input: Composition,
    mask: Tensor,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    out_component_tensor = torch.masked_select(
        input._component_tensor,
        mask[None],
        out=out._component_tensor if out is not None else None,
    ).reshape(input.numc(), -1)
    if out is not None:
        out._component_tensor = out._component_tensor.reshape(input.numc(), -1)
    out_residual_tensor = torch.masked_select(
        input._residual_tensor,
        mask,
        out=out._residual_tensor if out is not None else None,
    )
    return as_composition(out_component_tensor, out_residual_tensor)


@_auto_registration
def masked_scatter(input: Composition, mask: Tensor, source: Tensor) -> Composition:
    out_component_tensor = input._component_tensor.masked_scatter(mask[None], source)
    out_residual_tensor = input._residual_tensor.masked_scatter(mask, source)
    return as_composition(out_component_tensor, out_residual_tensor)


@overload
def gather(
    input: Composition,
    dim: _int,
    index: Tensor,
    *,
    sparse_grad: _bool = False,
    out: Optional[Composition] = None,
) -> Composition:
    ...


# TODO: to support 'out: Optional[Composition] = None'
@_auto_registration
def gather(
    input: Composition,
    dim: Any,
    index: Tensor,
    *,
    sparse_grad: _bool = False,
    out: Optional[Composition] = None,
) -> Composition:
    c_index = index[None].expand((input.numc(),) + (-1,) * index.dim())
    out_component_tensor = torch.gather(
        input._component_tensor,
        _shift_dim(dim),
        c_index,
        sparse_grad=sparse_grad,
        out=out._component_tensor if out is not None else None,
    )
    out_residual_tensor = torch.gather(
        input._residual_tensor,
        dim,
        index,
        sparse_grad=sparse_grad,
        out=out._residual_tensor if out is not None else None,
    )
    return as_composition(out_component_tensor, out_residual_tensor)


@overload
def scatter(
    input: Composition,
    dim: _int,
    index: Tensor,
    src: Tensor,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def scatter(
    input: Composition,
    dim: _int,
    index: Tensor,
    src: Tensor,
    *,
    reduce: str,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def scatter(
    input: Composition,
    dim: _int,
    index: Tensor,
    value: Number,
    *,
    reduce: str,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@overload
def scatter(
    input: Composition,
    dim: _int,
    index: Tensor,
    value: Number,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@_auto_registration
def scatter(
    input: Composition,
    dim: _int,
    index: Tensor,
    src_value: Optional[Union[Tensor, Number]] = None,
    *,
    reduce: Optional[str] = None,
    out: Optional[Composition] = None,
    src: Optional[Tensor] = None,
    value: Optional[Number] = None,
) -> Composition:
    r"""
    Unsafe.
    Safe when reduce is not None.
    """
    if src_value is None:
        src_value = src if src is not None else value
    assert src_value is not None
    if reduce == "add":
        holder = torch.zeros_like(input._residual_tensor).to(input._residual_tensor)
        holder = holder.scatter(dim, index, src_value, reduce=reduce)
        c_out = input + holder
        if out is not None:
            # TODO: use the out argument of `torch.add` raises an error
            out._component_tensor[:] = c_out._component_tensor
            out._residual_tensor[:] = c_out._residual_tensor
        return c_out
    else:
        c_index = index[None].expand((input.numc(),) + (-1,) * index.dim())
        c_src: Union[Tensor, Number]
        if isinstance(src_value, Tensor):
            c_src = src_value[None].expand((input.numc(),) + (-1,) * src_value.dim())
        else:
            c_src = src_value
        if reduce is None:
            out_component_tensor = torch.scatter(
                input._component_tensor,
                _shift_dim(dim),
                c_index,
                c_src,
                out=out._component_tensor if out is not None else None,
            )
            out_residual_tensor = torch.scatter(
                input._residual_tensor,
                dim,
                index,
                src_value,
                out=out._residual_tensor if out is not None else None,
            )
        else:
            out_component_tensor = torch.scatter(
                input._component_tensor,
                _shift_dim(dim),
                c_index,
                c_src,
                reduce=reduce,
                out=out._component_tensor if out is not None else None,
            )
            out_residual_tensor = torch.scatter(
                input._residual_tensor,
                dim,
                index,
                src_value,
                reduce=reduce,
                out=out._residual_tensor if out is not None else None,
            )
        return as_composition(out_component_tensor, out_residual_tensor)


@_auto_registration
def diagonal_scatter(
    input: Composition,
    src: Tensor,
    offset: _int = 0,
    dim1: _int = 0,
    dim2: _int = 1,
) -> Composition:
    if (
        torch.__version__ < "1.11.0"
    ):  # for versions < 1.11.0, 'diagonal_scatter' does not exist.
        raise RuntimeError("`diagonal_scatter` requires a torch version >= 1.11.0")
    c_src = src[None].expand((input.numc(),) + (-1,) * src.dim())
    out_component_tensor = input._component_tensor.diagonal_scatter(
        c_src, offset, _shift_dim(dim1), _shift_dim(dim2)
    )
    out_residual_tensor = input._residual_tensor.diagonal_scatter(
        src, offset, dim1, dim2
    )
    return as_composition(out_component_tensor, out_residual_tensor)


@overload
def index_select(
    input: Composition,
    dim: _int,
    index: Tensor,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@_auto_registration
def index_select(
    input: Composition,
    dim: _int,
    index: Tensor,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    out_component_tensor = torch.index_select(
        input._component_tensor,
        dim=_shift_dim(dim),
        index=index,
        out=out._component_tensor if out is not None else None,
    )
    out_residual_tensor = torch.index_select(
        input._residual_tensor,
        dim=dim,
        index=index,
        out=out._residual_tensor if out is not None else None,
    )
    return as_composition(out_component_tensor, out_residual_tensor)


@overload
def c_index_select(
    input: Composition,
    index: Tensor,
    with_residual: _bool = True,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    ...


def c_index_select(
    input: Composition,
    index: Tensor,
    with_residual: _bool = True,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    out_component_tensor = torch.index_select(
        input._component_tensor,
        dim=0,
        index=index,
        out=out._component_tensor if out is not None else None,
    )
    if with_residual:
        if out is not None:
            out._residual_tensor = out._residual_tensor.reshape_as(
                input._residual_tensor
            )
            out._residual_tensor[:] = input._residual_tensor
        out_residual_tensor = input._residual_tensor.clone()
    else:
        out_residual_tensor = torch.zeros_like(input._residual_tensor).to(
            input._residual_tensor
        )
    return as_composition(out_component_tensor, out_residual_tensor)


@_auto_registration
def masked_select(
    input: Composition,
    mask: Tensor,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    out_component_tensor = torch.masked_select(
        input._component_tensor,
        mask=mask[None],
        out=out._component_tensor if out is not None else None,
    ).reshape(input.numc(), -1)
    out_residual_tensor = torch.masked_select(
        input._residual_tensor,
        mask=mask,
        out=out._residual_tensor if out is not None else None,
    )
    if out is not None:
        out._component_tensor = out._component_tensor.reshape(input.numc(), -1)
    return as_composition(out_component_tensor, out_residual_tensor)


@overload
def select(input: Composition, dim: _int, index: _int) -> Composition:
    ...


def select(input: Composition, dim: Any, index: _int) -> Composition:
    out_component_tensor = torch.select(
        input._component_tensor, dim=_shift_dim(dim), index=index
    )
    out_residual_tensor = torch.select(input._residual_tensor, dim=dim, index=index)
    return as_composition(out_component_tensor, out_residual_tensor)


@overload
def index_fill(
    input: Composition, dim: _int, index: Tensor, value: Tensor
) -> Composition:
    ...


@overload
def index_fill(
    input: Composition, dim: _int, index: Tensor, value: Number
) -> Composition:
    ...


@_auto_registration
def index_fill(input: Composition, dim: _int, index: Tensor, value: Any) -> Composition:
    out_component_tensor = input._component_tensor.index_fill(
        dim=_shift_dim(dim), index=index, value=value
    )
    out_residual_tensor = input._residual_tensor.index_fill(
        dim=dim, index=index, value=value
    )
    return as_composition(out_component_tensor, out_residual_tensor)


@overload
def round(input: Composition, *, out: Optional[Composition] = None) -> Composition:
    ...


@overload
def round(
    input: Composition,
    *,
    decimals: _int,
    out: Optional[Composition] = None,
) -> Composition:
    ...


@_auto_registration
def round(
    input: Composition,
    *,
    decimals: _int = None,
    out: Optional[Composition] = None,
) -> Composition:
    if decimals is not None:
        out_component_tensor = torch.round(
            input._component_tensor,
            decimals=decimals,
            out=out._component_tensor if out is not None else None,
        )
        out_residual_tensor = torch.round(
            input._residual_tensor,
            decimals=decimals,
            out=out._residual_tensor if out is not None else None,
        )
    else:
        out_component_tensor = torch.round(input._component_tensor)
        out_residual_tensor = torch.round(input._residual_tensor)
    return as_composition(out_component_tensor, out_residual_tensor)


@overload
def round_(input: Composition) -> Composition:
    ...


@overload
def round_(input: Composition, *, decimals: _int) -> Composition:
    ...


@_auto_registration
def round_(input: Composition, *, decimals: _int = None) -> Composition:
    if decimals is not None:
        input._component_tensor.round_(decimals=decimals)
        input._residual_tensor.round_(decimals=decimals)
    else:
        input._component_tensor.round_()
        input._residual_tensor.round_()
    return input


@overload
def zeros(
    size: _size,
    c_num: _int,
    *,
    out: Optional[Composition] = None,
    dtype: _dtype = None,
    layout: Optional[_layout] = strided,
    device: Union[_device, str, None] = None,
    pin_memory: _bool = False,
    requires_grad: _bool = False,
) -> Composition:
    ...


@overload
def zeros(
    *size: _int,
    c_num: _int,
    out: Optional[Composition] = None,
    dtype: _dtype = None,
    layout: Optional[_layout] = strided,
    device: Union[_device, str, None] = None,
    pin_memory: _bool = False,
    requires_grad: _bool = False,
) -> Composition:
    ...


def zeros(*args: Any, **kwargs: Any):
    def _zeros(
        *,
        size: _size,
        c_num: _int,
        out: Optional[Composition] = None,
        dtype: _dtype = None,
        layout: Optional[_layout] = strided,
        device: Union[_device, str, None] = None,
        pin_memory: _bool = False,
        requires_grad: _bool = False,
    ):
        out_component_tensor = torch.zeros(
            (c_num,) + size,
            out=out._component_tensor if out is not None else None,
            dtype=dtype,
            layout=layout,
            device=device,
            pin_memory=pin_memory,
            requires_grad=requires_grad,
        )
        out_residual_tensor = torch.zeros(
            size,
            out=out._residual_tensor if out is not None else None,
            dtype=dtype,
            layout=layout,
            device=device,
            pin_memory=pin_memory,
            requires_grad=requires_grad,
        )
        return as_composition(out_component_tensor, out_residual_tensor)

    # parse args
    if len(args) > 0:
        if isinstance(args[0], _int):
            kwargs["size"] = args
        else:
            for i in range(len(args)):
                kwargs[["size", "c_num"][i]] = args[i]
    return _zeros(**kwargs)


@_auto_registration
def zeros_like(
    input: Composition,
    *,
    memory_format: Optional[memory_format] = None,
    dtype: _dtype = None,
    layout: Optional[_layout] = strided,
    device: Union[_device, str, None] = None,
    pin_memory: _bool = False,
    requires_grad: _bool = False,
) -> Composition:
    # TODO: fix bug. Default: if None, defaults to the dtype of input.
    out_component_tensor = torch.zeros_like(
        input._component_tensor,
        memory_format=memory_format,
        dtype=dtype,
        layout=layout,
        device=device,
        pin_memory=pin_memory,
        requires_grad=requires_grad,
    )
    out_residual_tensor = torch.zeros_like(
        input._residual_tensor,
        memory_format=memory_format,
        dtype=dtype,
        layout=layout,
        device=device,
        pin_memory=pin_memory,
        requires_grad=requires_grad,
    )
    return as_composition(out_component_tensor, out_residual_tensor)


def empty_indices(
    size: _size,
    c_num: _int,
    *,
    dtype: _dtype = torch.long,
    device: Union[_device, str, None] = None,
) -> IndexComposition:
    ...
    # TODO
    if dtype not in [torch.long, torch.int]:
        raise RuntimeError(
            "Expected argument 'dtype' to have one of the following scalar types: {}; but got {} instead".format(
                [torch.long, torch.int], dtype
            )
        )

    out_component_tensor = torch.zeros(size=(c_num,) + size, dtype=dtype, device=device)
    out_residual_tensor = torch.zeros(size=size, dtype=dtype, device=device)
    out_component_tensor[:] = IndexComposition.MASK_NUM
    out_residual_tensor[:] = IndexComposition.MASK_NUM
    return IndexComposition(out_component_tensor, out_residual_tensor)


@_auto_registration
def abs(input: Composition, *, out: Optional[Composition] = None) -> Composition:
    # TODO: bug: c.abs() not eqal to c.c_sum().abs(); This API has not been documented.
    out_component_tensor = torch.abs(
        input._component_tensor,
        out=out._component_tensor if out is not None else None,
    )
    out_residual_tensor = torch.abs(
        input._residual_tensor,
        out=out._residual_tensor if out is not None else None,
    )
    return as_composition(out_component_tensor, out_residual_tensor)


@_auto_registration
def abs_(input: Composition) -> Composition:
    # TODO: bug: c.abs() not eqal to c.c_sum().abs()
    torch.abs_(input._component_tensor)
    torch.abs_(input._residual_tensor)
    return input


@_auto_registration
def relu(input: Composition, *, ref: Optional[Tensor] = None) -> Composition:
    return core.decOVF.relu(input, ref=ref)


@_auto_registration
def relu_(input: Composition, *, ref: Optional[Tensor] = None) -> Composition:
    return core.decOVF.relu_(input, ref=ref)


@_auto_registration
def tanh(
    input: Composition,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    return core.decOVF.tanh(input, out=out, ref=ref)


@_auto_registration
def tanh_(input: Composition, *, ref: Optional[Tensor] = None) -> Composition:
    return core.decOVF.tanh_(input, ref=ref)


@_auto_registration
def sigmoid(
    input: Composition,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    return core.decOVF.sigmoid(input, out=out, ref=ref)


@_auto_registration
def sigmoid_(input: Composition, *, ref: Optional[Tensor] = None) -> Composition:
    return core.decOVF.sigmoid_(input, ref=ref)


@_auto_registration
def _pack_padded_sequence(input: Composition, lengths: Tensor, batch_first: _bool):
    if batch_first:
        input = input.transpose(0, 1)
    packed_input = []
    batch_sizes = torch.zeros((input.size(0),), dtype=torch.long, device="cpu")
    pack_num = input.size(1)
    for i in range(input.size(0)):
        packed_input.append(input[i, :pack_num])
        batch_sizes[i] = pack_num
        pack_num -= (lengths == i + 1).sum().item()
    packed_input = pydec.cat(packed_input, dim=0)

    return packed_input, batch_sizes


@_auto_registration
def _pad_packed_sequence(
    packed_input: Composition,
    batch_sizes: Tensor,
    batch_first: _bool,
    padding_value: _float,
    max_seq_length: _int,
):
    padded_input = []
    batch_start = 0
    max_batch_size = batch_sizes[0]
    feature_size = packed_input.size(-1)
    lengths = torch.zeros(
        (max_batch_size,),
        dtype=torch.long,
        device=packed_input.device,
    )
    for i in range(max_seq_length):
        batch_size = 0 if i >= len(batch_sizes) else batch_sizes[i]
        if batch_size > 0:
            padded_input.append(packed_input[batch_start : batch_start + batch_size])
            lengths[:batch_size] += 1
            batch_start += batch_size
        if max_batch_size - batch_size > 0:
            padding = pydec.zeros(
                (max_batch_size - batch_size, feature_size),
                c_num=packed_input.numc(),
                dtype=packed_input.dtype,
                device=packed_input.device,
            )
            if padding_value != 0:
                padding.residual[:] = padding_value
            padded_input.append(padding)
    padded_input = pydec.cat(padded_input, dim=0)
    padded_input = padded_input.view(-1, max_batch_size, feature_size)

    if batch_first:
        padded_input = padded_input.transpose_(0, 1)
    return padded_input, lengths


def _rnn_cell(
    input: Composition,
    hx: Union[Tensor, Composition],
    weight_ih: Parameter,
    weight_hh: Parameter,
    bias_ih: Parameter = None,
    bias_hh: Parameter = None,
    nonlinearity: str = "tanh",
    *,
    ref: Optional[Tensor] = None,
):
    if nonlinearity == "tanh":
        activation = tanh
    elif nonlinearity == "relu":
        activation = relu
    else:
        raise ValueError("Unknown nonlinearity '{}'".format(nonlinearity))
    return activation(
        pydec.nn.functional.linear(input, weight_ih, bias_ih)
        + pydec.nn.functional.linear(hx, weight_hh, bias_hh),
        ref=ref,
    )


def _rnn_layer(
    input: Composition,
    batch_sizes: Optional[Tensor],
    hx: Tensor,
    weight_ih: Parameter,
    weight_hh: Parameter,
    bias_ih: Parameter,
    bias_hh: Parameter,
    reverse: bool = False,
    nonlinearity: str = "tanh",
):
    """
    Could optimize performance by dropping completed samples
    """
    orig_hx = hx
    out = []
    out_hx = []
    batch_start = 0 if not reverse else input.size(0)
    if reverse:
        hx = None

    for i in (
        range(len(batch_sizes)) if not reverse else range(len(batch_sizes) - 1, -1, -1)
    ):
        batch_size = batch_sizes[i]
        if reverse:
            batch_size_new = batch_sizes[i]
            batch_size_old = 0 if i + 1 == len(batch_sizes) else batch_sizes[i + 1]
            if batch_size_new - batch_size_old > 0:
                if hx is None:
                    hx = orig_hx[batch_size_old:batch_size_new]
                else:
                    append_hx = pydec.zeros(
                        orig_hx[batch_size_old:batch_size_new].size(),
                        c_num=hx.numc(),
                        dtype=hx.dtype,
                        device=hx.device,
                    ).to(hx)
                    append_hx.residual[:] = orig_hx[batch_size_old:batch_size_new]
                    hx = pydec.cat(
                        [hx, append_hx],
                        dim=0,
                    )
        if not reverse:
            batch = input[batch_start : batch_start + batch_size]
            batch_start += batch_size
        else:
            batch = input[batch_start - batch_size : batch_start]
            batch_start -= batch_size
        hx = _rnn_cell(
            batch,
            hx,
            weight_ih,
            weight_hh,
            bias_ih,
            bias_hh,
            nonlinearity,
        )
        out.append(hx)
        if not reverse:
            batch_size_new = batch_sizes[i + 1] if i + 1 < len(batch_sizes) else 0
            batch_size_old = batch_sizes[i]
            if batch_size_old - batch_size_new > 0:
                out_hx.append(hx[batch_size_new:batch_size_old])
                hx = hx[:batch_size_new]
    if reverse:
        out = out[::-1]
        out_hx = hx
    else:
        out_hx = out_hx[::-1]
        out_hx = pydec.cat(out_hx, dim=0)
    out = pydec.cat(out, dim=0)
    return out, out_hx


def _rnn_packed(
    input: Composition,
    batch_sizes: Tensor,
    hx: Tensor,
    flat_weights: List[Parameter],
    bias: bool,
    num_layers: int,
    dropout: float,
    training: bool,
    bidirectional: bool,
    nonlinearity: str = "tanh",
):
    x = input
    weight_group_len = 2 * (2 if bias else 1) * (2 if bidirectional else 1)
    out_hx_list = []

    for i in range(num_layers):
        hx_layer, hx_layer_r = None, None
        weight_ih, weight_hh, bias_ih, bias_hh = (
            None,
            None,
            None,
            None,
        )
        weight_ih_r, weight_hh_r, bias_ih_r, bias_hh_r = (
            None,
            None,
            None,
            None,
        )
        weight_group = flat_weights[weight_group_len * i : weight_group_len * (i + 1)]

        weight_ih = weight_group[0]
        weight_hh = weight_group[1]
        if bias:
            bias_ih = weight_group[2]
            bias_hh = weight_group[3]
        if not bidirectional:
            hx_layer = hx[i]
            x, out_hx = _rnn_layer(
                x,
                batch_sizes,
                hx_layer,
                weight_ih,
                weight_hh,
                bias_ih,
                bias_hh,
                nonlinearity=nonlinearity,
            )
            out_hx_list.append(out_hx)
        else:
            weight_group = weight_group[len(weight_group) // 2 :]  # takes the back half
            hx_layer = hx[2 * i]
            hx_layer_r = hx[2 * i + 1]
            weight_ih_r = weight_group[0]
            weight_hh_r = weight_group[1]
            if bias:
                bias_ih_r = weight_group[2]
                bias_hh_r = weight_group[3]

            x_, out_hx = _rnn_layer(
                x,
                batch_sizes,
                hx_layer,
                weight_ih,
                weight_hh,
                bias_ih,
                bias_hh,
                nonlinearity=nonlinearity,
            )
            x_r, out_hx_r = _rnn_layer(
                x,
                batch_sizes,
                hx_layer_r,
                weight_ih_r,
                weight_hh_r,
                bias_ih_r,
                bias_hh_r,
                True,
                nonlinearity=nonlinearity,
            )
            x = pydec.cat([x_, x_r], dim=-1)
            out_hx_list.append(out_hx)
            out_hx_list.append(out_hx_r)
        if training and dropout > 0:
            # TODO: should also apply dropout to out_hx?
            x = pydec.nn.functional.dropout(
                x, p=dropout, training=training, inplace=True
            )

    out = x
    out_hx = pydec.stack(out_hx_list, dim=0)
    return out, out_hx


@overload
def rnn_relu(
    data: Composition,
    batch_sizes: Tensor,
    hx: Tensor,
    params: Union[Tuple[Tensor, ...], List[Tensor]],
    has_biases: _bool,
    num_layers: _int,
    dropout: _float,
    train: _bool,
    bidirectional: _bool,
    /,
) -> Tuple[Composition, Composition]:
    ...


@overload
def rnn_relu(
    input: Composition,
    hx: Tensor,
    params: Union[Tuple[Tensor, ...], List[Tensor]],
    has_biases: _bool,
    num_layers: _int,
    dropout: _float,
    train: _bool,
    bidirectional: _bool,
    batch_first: _bool,
    /,
) -> Tuple[Composition, Composition]:
    ...


@_auto_registration
def rnn_relu(
    *args,
):
    """
    Only supports the positional arguments in current version
    """
    if isinstance(args[3], bool):
        (
            input,
            hx,
            params,
            has_biases,
            num_layers,
            dropout,
            train,
            bidirectional,
            batch_first,
        ) = args
        batch_sizes = None
        if batch_first:
            input: Composition = input.transpose(0, 1)
        seq_len = input.size(0)
        batch_size = input.size(1)
        batch_sizes = torch.tensor(
            [batch_size] * input.size(0),
            dtype=torch.long,
            device="cpu",
        )
        input = input.view(seq_len * batch_size, -1)
        out, out_hx = _rnn_packed(
            input,
            batch_sizes,
            hx,
            params,
            has_biases,
            num_layers,
            dropout,
            train,
            bidirectional,
            nonlinearity="relu",
        )
        out = out.view(seq_len, batch_size, -1)
        if batch_first:
            out.transpose_(0, 1)
        return out, out_hx
    else:
        (
            data,
            batch_sizes,
            hx,
            params,
            has_biases,
            num_layers,
            dropout,
            train,
            bidirectional,
        ) = args
        return _rnn_packed(
            data,
            batch_sizes,
            hx,
            params,
            has_biases,
            num_layers,
            dropout,
            train,
            bidirectional,
            nonlinearity="relu",
        )


@overload
def rnn_tanh(
    data: Composition,
    batch_sizes: Tensor,
    hx: Tensor,
    params: Union[Tuple[Tensor, ...], List[Tensor]],
    has_biases: _bool,
    num_layers: _int,
    dropout: _float,
    train: _bool,
    bidirectional: _bool,
    /,
) -> Tuple[Composition, Composition]:
    ...


@overload
def rnn_tanh(
    input: Composition,
    hx: Tensor,
    params: Union[Tuple[Tensor, ...], List[Tensor]],
    has_biases: _bool,
    num_layers: _int,
    dropout: _float,
    train: _bool,
    bidirectional: _bool,
    batch_first: _bool,
    /,
) -> Tuple[Composition, Composition]:
    ...


@_auto_registration
def rnn_tanh(
    *args,
):
    """
    Only supports the positional arguments in current version
    """
    if isinstance(args[3], bool):
        (
            input,
            hx,
            params,
            has_biases,
            num_layers,
            dropout,
            train,
            bidirectional,
            batch_first,
        ) = args
        batch_sizes = None
        if batch_first:
            input: Composition = input.transpose(0, 1)
        seq_len = input.size(0)
        batch_size = input.size(1)
        batch_sizes = torch.tensor(
            [batch_size] * input.size(0),
            dtype=torch.long,
            device="cpu",
        )
        input = input.view(seq_len * batch_size, -1)
        out, out_hx = _rnn_packed(
            input,
            batch_sizes,
            hx,
            params,
            has_biases,
            num_layers,
            dropout,
            train,
            bidirectional,
            nonlinearity="tanh",
        )
        out = out.view(seq_len, batch_size, -1)
        if batch_first:
            out.transpose_(0, 1)
        return out, out_hx
    else:
        (
            data,
            batch_sizes,
            hx,
            params,
            has_biases,
            num_layers,
            dropout,
            train,
            bidirectional,
        ) = args
        return _rnn_packed(
            data,
            batch_sizes,
            hx,
            params,
            has_biases,
            num_layers,
            dropout,
            train,
            bidirectional,
            nonlinearity="tanh",
        )


@_auto_registration
def reciprocal(
    input: Composition,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    return core.decOVF.reciprocal(input, out=out, ref=ref)


@_auto_registration
def reciprocal_(
    input: Composition,
    *,
    ref: Optional[Tensor] = None,
) -> Composition:
    return core.decOVF.reciprocal_(input, ref=ref)


@_auto_registration
def exp(
    input: Composition,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    return core.decOVF.exp(input, out=out, ref=ref)


@_auto_registration
def exp_(
    input: Composition,
    *,
    ref: Optional[Tensor] = None,
) -> Composition:
    return core.decOVF.exp_(input, ref=ref)


@overload
def softmax(
    input: Composition,
    dim: _int,
    dtype: Optional[_dtype] = None,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    ...


@_auto_registration
def softmax(
    input: Composition,
    dim: _int,
    dtype: Optional[_dtype] = None,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    # TODO: should we disable grad here?
    # TODO: This API has not been documented.
    if dtype is not None:
        input = input.to(dtype)
        if ref is not None:
            ref = ref.to(dtype)
    input = _softmax_rescale(input)
    components_bias = torch.max(input.components, dim=0, keepdim=False)[0]
    components_bias = -torch.max(components_bias, dim=dim, keepdim=True)[0]
    residual_bias = -torch.max(input.residual, dim=dim, keepdim=True)[0]
    bias = torch.min(components_bias, residual_bias)
    input = core.decOVF.biased_exp(input, bias=bias, ref=ref)
    if ref is None:
        exp_sum = input.c_sum().sum(dim=dim, keepdim=True)
    else:
        exp_sum = torch.exp(ref + bias)
    return pydec.div(input, exp_sum, out=out)


@_auto_registration
def sqrt(
    input: Composition,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    return core.decOVF.sqrt(input, out=out, ref=ref)


@_auto_registration
def sqrt_(
    input: Composition,
    *,
    ref: Optional[Tensor] = None,
) -> Composition:
    return core.decOVF.sqrt_(input, ref=ref)


@overload
def var(
    input: Composition,
    dim: Union[_int, _size],
    unbiased: _bool = True,
    keepdim: _bool = False,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    ...


@overload
def var(
    input: Composition,
    unbiased: _bool = True,
    *,
    ref: Optional[Tensor] = None,
) -> Composition:
    ...


@_auto_registration
def var(
    input: Composition,
    dim: Union[_int, _size] = None,
    unbiased: _bool = True,
    keepdim: _bool = False,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    if isinstance(dim, _bool):
        unbiased = dim
        dim = None
    if dim is None:
        dim = tuple(range(0, input.dim()))
    return core.decMVF.var(input, dim, unbiased, keepdim, out=out, ref=ref)


@_auto_registration
def square(
    input: Composition,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    return core.decOVF.square(input, out=out, ref=ref)


@_auto_registration
def square_(
    input: Composition,
    ref: Optional[Tensor] = None,
) -> Composition:
    return core.decOVF.square_(input, ref=ref)


def _softmax_rescale(input: Composition) -> Composition:
    """
    This is not a standard api of PyTorch.
    Use to deal with numerical explosion problem in decomposition.
    """
    combine_components = torch.cat(
        (
            input.components,
            input.residual[None],
        )
    )
    softmax_components = torch.softmax(combine_components, dim=0)
    rescaled_components = softmax_components * input.c_sum()
    out_components = rescaled_components[:-1]
    out_residual = rescaled_components[-1]
    return as_composition(out_components, out_residual)


@overload
def eq(input: Composition, other: Tensor, *, out: Optional[Tensor] = None) -> Tensor:
    ...


@overload
def eq(
    input: Composition, other: Composition, *, out: Optional[Tensor] = None
) -> Tensor:
    ...


@overload
def eq(input: Composition, other: Number, *, out: Optional[Tensor] = None) -> Tensor:
    ...


def eq(input: Composition, other: Any, *, out: Optional[Tensor] = None) -> Tensor:
    if isinstance(other, Composition):
        return input.c_sum().eq(other.c_sum())
    return input.c_sum().eq(other)


@overload
def ne(input: Composition, other: Tensor, *, out: Optional[Tensor] = None) -> Tensor:
    ...


@overload
def ne(
    input: Composition, other: Composition, *, out: Optional[Tensor] = None
) -> Tensor:
    ...


@overload
def ne(input: Composition, other: Number, *, out: Optional[Tensor] = None) -> Tensor:
    ...


def ne(input: Composition, other: Any, *, out: Optional[Tensor] = None) -> Tensor:
    if isinstance(other, Composition):
        return input.c_sum().ne(other.c_sum())
    return input.c_sum().ne(other)


@overload
def gt(input: Composition, other: Tensor, *, out: Optional[Tensor] = None) -> Tensor:
    ...


@overload
def gt(
    input: Composition, other: Composition, *, out: Optional[Tensor] = None
) -> Tensor:
    ...


@overload
def gt(input: Composition, other: Number, *, out: Optional[Tensor] = None) -> Tensor:
    ...


def gt(input: Composition, other: Any, *, out: Optional[Tensor] = None) -> Tensor:
    if isinstance(other, Composition):
        return input.c_sum().gt(other.c_sum())
    return input.c_sum().gt(other)


@overload
def lt(input: Composition, other: Tensor, *, out: Optional[Tensor] = None) -> Tensor:
    ...


@overload
def lt(
    input: Composition, other: Composition, *, out: Optional[Tensor] = None
) -> Tensor:
    ...


@overload
def lt(input: Composition, other: Number, *, out: Optional[Tensor] = None) -> Tensor:
    ...


def lt(input: Composition, other: Any, *, out: Optional[Tensor] = None) -> Tensor:
    if isinstance(other, Composition):
        return input.c_sum().lt(other.c_sum())
    return input.c_sum().lt(other)


@overload
def ge(input: Composition, other: Tensor, *, out: Optional[Tensor] = None) -> Tensor:
    ...


@overload
def ge(
    input: Composition, other: Composition, *, out: Optional[Tensor] = None
) -> Tensor:
    ...


@overload
def ge(input: Composition, other: Number, *, out: Optional[Tensor] = None) -> Tensor:
    ...


def ge(input: Composition, other: Any, *, out: Optional[Tensor] = None) -> Tensor:
    if isinstance(other, Composition):
        return input.c_sum().ge(other.c_sum())
    return input.c_sum().ge(other)


@overload
def le(input: Composition, other: Tensor, *, out: Optional[Tensor] = None) -> Tensor:
    ...


@overload
def le(
    input: Composition, other: Composition, *, out: Optional[Tensor] = None
) -> Tensor:
    ...


@overload
def le(input: Composition, other: Number, *, out: Optional[Tensor] = None) -> Tensor:
    ...


def le(input: Composition, other: Any, *, out: Optional[Tensor] = None) -> Tensor:
    if isinstance(other, Composition):
        return input.c_sum().le(other.c_sum())
    return input.c_sum().le(other)



from __future__ import annotations
import functools
import inspect

from torch.autograd.grad_mode import _DecoratorContextManager

from typing import Dict, Tuple, Union, Any, Callable, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ..._composition import Composition

from torch.types import (
    _int,
    _float,
    _bool,
    Number,
    _dtype,
    _device,
    _qscheme,
    _size,
    _layout,
)

_DECOMPOSITION_FUNC_REGISTRY = {}


class _DecompositionState:
    decomposition_name: str = None
    decomposition_args: Dict[str, Any] = {}


def register_decomposition_func(name: str):
    """
    TODO: docs need update
    New decomposition_func can be added with the :func:`register_decomposition_func`
    function decorator.

    For example::

        @register_model('forward_norm_decomposition')
        def abs_decomposition(input: Composition, func: Callable[[Tensor], Tensor], *, ref: Optional[Tensor] = None, inplace: _bool = False,):
            (...)

    Args:
        name (str): the name of the funcion
    """

    def register_func(func):
        if name in _DECOMPOSITION_FUNC_REGISTRY:
            raise ValueError("Cannot register duplicate function ({})".format(name))
        # if name == "none":
        #     raise ValueError(
        #         'Cannot register function ({}), the name "none" is reserved.'.format(
        #             name
        #         )
        #     )

        @functools.wraps(func)
        def warp_decomposition_func(*args, **kwargs):
            dec_args = _DecompositionState.decomposition_args.copy()
            dec_args.update(kwargs)
            argspec = inspect.getfullargspec(func)
            if argspec.varkw is None:
                ignore_keys = []
                for key in dec_args.keys():
                    if key not in (argspec.args + argspec.kwonlyargs):
                        ignore_keys.append(key)
                for key in ignore_keys:
                    dec_args.pop(key)
            return func(*args, **dec_args)

        _DECOMPOSITION_FUNC_REGISTRY[name] = warp_decomposition_func

        return warp_decomposition_func

    return register_func


def set_decomposition_func(name: str) -> None:
    if name not in _DECOMPOSITION_FUNC_REGISTRY:
        raise ValueError("decomposition function ({}) is not registered".format(name))
    _DecompositionState.decomposition_name = name


def get_decomposition_name() -> str:
    return _DecompositionState.decomposition_name


def get_decomposition_func() -> Callable[..., Composition]:
    if _DecompositionState.decomposition_name not in _DECOMPOSITION_FUNC_REGISTRY:
        return None
    else:
        current_decomposition_func = _DECOMPOSITION_FUNC_REGISTRY[
            _DecompositionState.decomposition_name
        ]
        return current_decomposition_func


class using_decomposition_func(_DecoratorContextManager):
    def __init__(self, name: str) -> None:
        if name not in _DECOMPOSITION_FUNC_REGISTRY:
            raise ValueError(
                "Decomposition function ({}) is not registered".format(name)
            )
        self.prev = None
        self.using_name = name

    def __enter__(self):
        self.prev = _DecompositionState.decomposition_name
        _DecompositionState.decomposition_name = self.using_name

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        _DecompositionState.decomposition_name = self.prev

    def clone(self):
        return self.__class__(self.using_name)


def set_decomposition_args(update: _bool = True, **kwargs) -> None:
    if update:
        _DecompositionState.decomposition_args.update(kwargs)
    else:
        _DecompositionState.decomposition_args = kwargs


def get_decomposition_args() -> Dict[str, Any]:
    return _DecompositionState.decomposition_args


class using_decomposition_args(_DecoratorContextManager):
    def __init__(self, update: _bool = True, **kwargs) -> None:
        self.update = update
        self.prev = None
        self.using_args = kwargs

    def __enter__(self):
        self.prev = _DecompositionState.decomposition_args.copy()
        if self.update:
            _DecompositionState.decomposition_args.update(self.using_args)
        else:
            _DecompositionState.decomposition_args = self.using_args

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        _DecompositionState.decomposition_args = self.prev

    def clone(self):
        return self.__class__(self.update, **self.using_args)



"""
Decomposition of Multi-Variable Functions.
"""

from __future__ import annotations
import torch
import pydec
from torch import Tensor

from ...utils import parse_args

from typing import (
    Dict,
    List,
    Tuple,
    Union,
    Any,
    Callable,
    Optional,
    overload,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from ..._composition import Composition

from torch.types import (
    _int,
    _float,
    _bool,
    Number,
    _dtype,
    _device,
    _qscheme,
    _size,
    _layout,
)

import functools

from pydec.exception_utils import args_error

# TODO: disable grad tracing in ops


# def _decomposeMVF_grad(
#     input: Composition, out_cbody: Tensor, out_residual: Tensor, grad: Tensor
# ) -> Composition:
#     if (
#         not isinstance(input, Composition)
#         or not isinstance(out_cbody, Tensor)
#         or not isinstance(out_residual, Tensor)
#         or not isinstance(grad, Tensor)
#     ):
#         raise args_error(decomposeMVF.__name__, input, out_cbody, out_residual, grad)
#     out_components = torch.matmul(input.components, grad.transpose(-1, -2))
#     multiplier = out_cbody / out_components.sum(dim=0)
#     out_components *= multiplier
#     return pydec.as_composition(out_components, out_residual)


# def _decomposeMVF_func(
#     input: Composition,
#     func: Callable[[Tensor], Tensor],
#     ref: Optional[Tensor] = None,
#     grad: Optional[Tensor] = None,
# ):
#     if (
#         not isinstance(input, Composition)
#         or (ref is not None and not isinstance(ref, Tensor))
#         or (grad is not None and not isinstance(grad, Tensor))
#     ):
#         raise args_error(decomposeMVF.__name__, input, func, ref, grad)
#     if ref is None:
#         ref = input.c_sum()
#     if grad is not None:
#         with torch.no_grad():
#             c_out = func(ref)
#             out_residual = func(input.residual)
#             out_cbody = c_out - out_residual
#             out_dim = c_out.size(-1)
#             in_dim = input.size(-1)
#             if grad.dim() < 2 or grad.size()[-2:] != (out_dim, in_dim):
#                 raise RuntimeError  # TODO: add msg
#     else:
#         input_grad_holder = ref.clone().detach_()
#         input_grad_holder.requires_grad_(True)
#         with torch.enable_grad():
#             c_out = func(input_grad_holder)
#             grad = []
#             for i in range(c_out.size(-1)):
#                 ...
#             # loss = c_out.sum(dim=0,...,-2)
#             # TODO
#     return _decomposeMVF_grad(input, out_cbody, out_residual, grad, inplace=inplace)


# @overload
# def decomposeMVF(
#     input: Composition,
#     func: Callable[[Tensor], Tensor],
#     ref: Optional[Tensor] = None,
#     grad: Optional[Tensor] = None,
# ) -> Composition:
#     ...


# @overload
# def decomposeMVF(
#     input: Composition, out_cbody: Tensor, out_residual: Tensor, grad: Tensor
# ) -> Composition:
#     ...


# def decomposeMVF(*args, **kwargs) -> Composition:
#     mode = "func"
#     if len(args) > 1:
#         if isinstance(args[1], Tensor):
#             mode = "grad"
#     elif "cbody" in kwargs:
#         mode = "grad"
#     if mode == "func":
#         parse_args(args, ["input", "func", "ref", "grad"], kwargs)
#         return _decomposeMVF_func(**kwargs)
#     else:
#         parse_args(args, ["input", "out_cbody", "out_residual", "grad"], kwargs)
#         return _decomposeMVF_grad(**kwargs)


__all__ = [
    "cc_mul",
    "cc_mul_",
    "cc_div",
    "cc_div_composite",
    "cc_matmul",
    "cc_bmm",
    "var",
]


def cc_mul(
    input: Composition,
    other: Composition,
    *,
    out: Optional[Composition] = None,
    ref_input: Optional[Tensor] = None,
    ref_other: Optional[Tensor] = None,
) -> Composition:
    """
    f(x1,x2) = x1*x2 = (c1+b1)*(c2+b2) = fc1 + fc2
    fc1 = 0.5*x1*x2 - 0.5*b1*x2 + 0.5*x1*b2 - 0.5*b1*b2
    fc2 = 0.5*x1*x2 - 0.5*x1*b2 + 0.5*b1*x2 - 0.5*b1*b2
    """
    if ref_input is None:
        ref_input = input.c_sum()
    if ref_other is None:
        ref_other = other.c_sum()
    out_residual = torch.mul(input.residual, other.residual)
    ref_out = torch.mul(ref_input, ref_other)
    shapley_bias = -torch.mul(input.residual, ref_other) + torch.mul(
        ref_input, other.residual
    )  # -b1*x2 + x1*b2

    c_body = ref_out - out_residual
    out_c1 = 0.5 * (c_body + shapley_bias)
    out_c2 = 0.5 * (c_body - shapley_bias)
    # out_c1 = 0.5 * c_body
    # out_c2 = out_c1

    # TODO: bug: for 0-dim tensor, should append a dim to composition (components tensor should have at least 2 dim)
    multiplier1 = out_c1 / input.components.sum(dim=0, keepdim=True)
    multiplier2 = out_c2 / other.components.sum(dim=0, keepdim=True)
    multiplier1.nan_to_num_(0, 0, 0)
    multiplier2.nan_to_num_(0, 0, 0)
    out_component_tensor = (
        multiplier1 * input._component_tensor + multiplier2 * other._component_tensor
    )
    if out is not None:
        out.residual[:] = out_residual
        out.components[:] = out_component_tensor
    return pydec.as_composition(out_component_tensor, out_residual)


def cc_mul_(
    input: Composition,
    other: Composition,
    *,
    ref_input: Optional[Tensor] = None,
    ref_other: Optional[Tensor] = None,
) -> Composition:
    raise NotImplementedError


def cc_div(
    input: Composition,
    other: Composition,
    *,
    rounding_mode: Optional[str] = None,
    out: Optional[Composition] = None,
    ref_input: Optional[Tensor] = None,
    ref_other: Optional[Tensor] = None,
) -> Composition:
    # return cc_div_composite(
    #     input, other, out=out, ref_input=ref_input, ref_other=ref_other
    # )
    other = pydec.reciprocal(other, ref=ref_other)
    return cc_mul(
        input,
        other,
        out=out,
        ref_input=ref_input,
        ref_other=ref_other,
    )


def cc_div_composite(
    input: Composition,
    other: Composition,
    *,
    rounding_mode: Optional[str] = None,
    out: Optional[Composition] = None,
    ref_input: Optional[Tensor] = None,
    ref_other: Optional[Tensor] = None,
) -> Composition:
    """
    f(x1,x2) = x1/x2 = (c1+b1)/(c2+b2) = fc1 + fc2
    fc1 = 0.5(x1/x2 - b1/x2 + x1/b2 - b1/b2)
    = 0.5[(x1-b1)/x2 + (x1-b1)/b2]
    fc2 = 0.5(x1/x2 - x1/b2 + b1/x2 - b1/b2)
    = 0.5[(x1+b1)/x2 - (x1+b1)/b2]

    TODO: not stable. (1+0)/(2+0) vs. (1+0)/(2+0.01)
    """
    if ref_input is None:
        ref_input = input.c_sum()
    if ref_other is None:
        ref_other = other.c_sum()
    out_residual = torch.div(
        input.residual, other.residual, rounding_mode=rounding_mode
    ).nan_to_num_(0, 0, 0)
    term00 = torch.div(
        ref_input - input.residual, ref_other, rounding_mode=rounding_mode
    ).nan_to_num_(0, 0, 0)
    term01 = torch.div(
        ref_input - input.residual, other.residual, rounding_mode=rounding_mode
    ).nan_to_num_(0, 0, 0)
    term10 = torch.div(
        ref_input + input.residual, ref_other, rounding_mode=rounding_mode
    ).nan_to_num_(0, 0, 0)
    term11 = torch.div(
        ref_input + input.residual, other.residual, rounding_mode=rounding_mode
    ).nan_to_num_(0, 0, 0)

    out_c1 = 0.5 * (term00 + term01)
    out_c2 = 0.5 * (term10 - term11)
    multiplier1 = out_c1 / input.components.sum(dim=0, keepdim=True)
    multiplier2 = out_c2 / other.components.sum(dim=0, keepdim=True)

    multiplier1.nan_to_num_(0, 0, 0)
    multiplier2.nan_to_num_(0, 0, 0)
    out_component_tensor = (
        multiplier1 * input._component_tensor + multiplier2 * other._component_tensor
    )
    if out is not None:
        out.residual[:] = out_residual
        out.components[:] = out_component_tensor
    return pydec.as_composition(out_component_tensor, out_residual)


def cc_matmul(
    input: Composition,
    other: Composition,
    *,
    out: Optional[Composition] = None,
    ref_input: Optional[Tensor] = None,
    ref_other: Optional[Tensor] = None,
) -> Composition:
    if input.dim() == 1:
        return cc_matmul(
            input.unsqueeze(0), other, out=out, ref_input=ref_input, ref_other=ref_other
        ).squeeze_(-2)
    if other.dim() == 1:
        return cc_matmul(
            input,
            other.unsqueeze(-1),
            out=out,
            ref_input=ref_input,
            ref_other=ref_other,
        ).squeeze_(-1)
    input = input.unsqueeze(-2)
    other = other.unsqueeze(-3).transpose(-1, -2)
    c_out = pydec.mul(input, other, ref_input=ref_input, ref_other=ref_other).sum(-1)
    if out is not None:
        out.components[:] = c_out.components
        out.residual[:] = c_out.residual
    return c_out


def var(
    input: Composition,
    dim: Union[_int, _size],
    unbiased: _bool = True,
    keepdim: _bool = False,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    # TODO: apply linearization
    if ref is None:
        ref = input.c_sum()
    input_mean = input.mean(dim=dim, keepdim=True)
    ref_mean = ref.mean(dim=dim, keepdim=True)

    input_diff = input - input_mean
    ref_diff = ref - ref_mean
    input_vars = pydec.square(input_diff, ref=ref_diff)
    if unbiased:
        if isinstance(dim, _int):
            dim = (dim,)
        unbiased_num = (
            functools.reduce(lambda x, y: x * y, [ref.size(d) for d in dim]) - 1
        )
        return pydec.div(
            pydec.sum(input_vars, dim=dim, keepdim=keepdim), unbiased_num, out=out
        )
    else:
        return pydec.mean(input_vars, dim=dim, keepdim=keepdim, out=out)


def cc_bmm(
    input: Composition,
    mat2: Composition,
    *,
    out: Optional[Composition] = None,
    ref_input: Optional[Tensor] = None,
    ref_mat2: Optional[Tensor] = None,
) -> Composition:
    input = input.unsqueeze(-2)  # b x n x 1 x m
    mat2 = mat2.unsqueeze(-3).transpose(-1, -2)  # b x 1 x p x m
    c_out = pydec.mul(input, mat2, ref_input=ref_input, ref_other=ref_mat2).sum(-1)
    if out is not None:
        out.components[:] = c_out.components
        out.residual[:] = c_out.residual
    return c_out


# deprecated
def layer_norm_loo(
    input: Composition,
    normalized_shape: List[int],
    weight: Optional[Tensor] = None,
    bias: Optional[Tensor] = None,
    eps: float = 1e-5,
    *,
    ref: Optional[Tensor] = None,
) -> Composition:
    normalized_dims = tuple(range(-len(normalized_shape), 0))
    input_mean = input.mean(dim=normalized_dims, keepdim=True)
    input = input - input_mean

    if ref is None:
        ref = input.c_sum()
    else:
        ref = ref - ref.mean(dim=normalized_dims, keepdim=True)

    def central_norm(
        input: Tensor,
        normalized_dims: Tuple[int, ...],
        eps: float = 1e-5,
    ):
        # return torch.nn.functional.layer_norm(input, [input.size(-1)])
        return input / torch.sqrt(
            torch.var(input, dim=normalized_dims, unbiased=False, keepdim=True) + eps
        )

    ref_out = central_norm(ref, normalized_dims, eps=eps)
    out_residual = central_norm(input.residual, normalized_dims, eps=eps)

    out_components = torch.zeros_like(input.components)
    for i in range(input.numc()):
        masked_input: Composition = pydec.as_composition(
            input.components.clone(), input.residual
        )
        masked_input()[i] = 0
        out_components[i] = ref_out - central_norm(
            masked_input.c_sum(), normalized_dims, eps=eps
        )
    scaler = (ref_out - out_residual) / out_components.sum(dim=0)
    out_components *= scaler

    out = pydec.as_composition(out_components, out_residual)
    if weight is not None:
        out *= weight
    if bias is not None:
        out += bias
    return out


# deprecated
def layer_norm_grad(
    input: Composition,
    normalized_shape: List[int],
    weight: Optional[Tensor] = None,
    bias: Optional[Tensor] = None,
    eps: float = 1e-5,
    *,
    ref: Optional[Tensor] = None,
) -> Composition:
    """
    step1: input = input - input.mean()
    step2: do mask

    diff = \frac{termb - terma}{terma*termb}x_i\\
    terma = \sqrt{\frac{1}{m}\sum_ix_i^2}\\
    termb = \sqrt{\frac{1}{m}\sum_ix_i^2-\frac{1}{m}x_j^2}
    """
    normalized_dims = tuple(range(-len(normalized_shape), 0))
    batch_shape = input.size()[: (input.ndim - len(normalized_shape))]
    input = input.view(batch_shape + (-1,))
    input = input - input.mean(dim=-1, keepdim=True)

    normalized_num = input.size(-1)

    if ref is None:
        ref = input.c_sum()
    else:
        ref.view(batch_shape + (-1,))
        ref = ref - ref.mean(dim=-1, keepdim=True)

    def central_norm(
        input: Tensor,
        normalized_dims: Tuple[int, ...],
        eps: float = 1e-5,
    ):
        # return torch.nn.functional.layer_norm(input, normalized_shape, eps=eps)
        return input / torch.sqrt(
            input.square().sum(dim=normalized_dims, keepdim=True) / input.size(-1) + eps
        )

    ref_out = central_norm(ref, (-1,), eps=eps)
    out_residual = central_norm(input.residual, (-1,), eps=eps)

    square_ref = ref.square()
    terma = square_ref.sum(dim=-1, keepdim=True)
    termb = terma - square_ref
    terma /= normalized_num
    terma.sqrt_()
    termb /= normalized_num
    termb.sqrt_()

    tw = ref.unsqueeze(-1) * ((termb - terma) / (termb * terma)).unsqueeze(-2)
    tw = tw.diagonal_scatter(ref_out, 0, -2, -1)

    # weight = torch.zeros((ref.size(0), ref.size(0)))
    # for i in range(len(weight)):
    #     new_ref = ref.clone()
    #     new_ref[i] = 0
    #     weight[:, i] = ref_out - central_norm(new_ref, [-1])

    # scaler = (ref_out - out_residual) / weight.sum(dim=1)
    # weight *= scaler[:, None]
    # weight /= input.components.sum(dim=0)
    # out_components = input.components @ weight.transpose(-1, -2)

    scaler = (ref_out - out_residual) / tw.sum(dim=-1)
    tw *= scaler.unsqueeze(-1)

    tw /= input.components.sum(dim=0).unsqueeze(-2)
    out_components = input.components @ tw.transpose(-1, -2)
    out = pydec.as_composition(out_components, out_residual)
    out = out.view(batch_shape + tuple(normalized_shape))

    if weight is not None:
        out *= weight
    if bias is not None:
        out += bias
    return out



from .meta import CopyModule
from .module import DecModule
import torch.nn as nn


class Container(nn.Container, DecModule, metaclass=CopyModule):
    __doc__ = nn.Container.__doc__


class Sequential(nn.Sequential, DecModule, metaclass=CopyModule):
    __doc__ = nn.Sequential.__doc__


class ModuleList(nn.ModuleList, DecModule, metaclass=CopyModule):
    __doc__ = nn.ModuleList.__doc__


class ModuleDict(nn.ModuleDict, DecModule, metaclass=CopyModule):
    __doc__ = nn.ModuleDict.__doc__


class ParameterList(nn.ParameterList, DecModule, metaclass=CopyModule):
    __doc__ = nn.ParameterList.__doc__


class ParameterDict(nn.ParameterDict, DecModule, metaclass=CopyModule):
    __doc__ = nn.ParameterDict.__doc__



from __future__ import annotations
import pydec
import torch
from torch import Tensor
from typing import Any, Union, List, Tuple, Optional, Callable, overload, TYPE_CHECKING

if TYPE_CHECKING:
    from pydec import Composition


# In some cases, these basic types are shadowed by corresponding
# top-level values.  The underscore variants let us refer to these
# types.  See https://github.com/python/mypy/issues/4146 for why these
# workarounds is necessary
# In the future, this will become useful if mypy is introduced into pydec
from torch.types import (
    _int,
    _float,
    _bool,
    Number,
    _dtype,
    _device,
    _qscheme,
    _size,
    _layout,
)

from pydec.exception_utils import component_num_error

__all__ = [
    "cc_add",
    "cc_add_",
    "ct_add",
    "ct_add_",
    "tc_add",
    "cc_sub",
    "cc_sub_",
    "ct_sub",
    "ct_sub_",
    "tc_sub",
    "ct_mul",
    "ct_mul_",
    "tc_mul",
    "ct_div",
    "ct_div_",
    "ct_matmul",
    "tc_matmul",
    "ct_mv",
    "tc_mv",
    "ct_mm",
    "tc_mm",
    "ct_bmm",
    "tc_bmm",
]


def cc_add(
    input: Composition,
    other: Composition,
    *,
    alpha: Optional[Number] = 1,
    out: Optional[Composition] = None,
) -> Composition:
    if input.numc() != other.numc():
        raise component_num_error(input.numc(), other.numc())
    out_component_tensor = torch.add(
        input._component_tensor,
        other._component_tensor,
        alpha=alpha,
        out=out._component_tensor if out is not None else None,
    )
    out_residual_tensor = torch.add(
        input._residual_tensor,
        other._residual_tensor,
        alpha=alpha,
        out=out._residual_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def cc_add_(
    input: Composition,
    other: Composition,
    *,
    alpha: Optional[Number] = 1,
) -> Composition:
    if input.numc() != other.numc():
        raise component_num_error(input.numc(), other.numc())
    out_component_tensor = input._component_tensor.add_(
        other._component_tensor,
        alpha=alpha,
    )
    out_residual_tensor = input._residual_tensor.add_(
        other._residual_tensor,
        alpha=alpha,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def ct_add(
    input: Composition,
    other: Union[Tensor, Number],
    *,
    alpha: Optional[Number] = 1,
    out: Optional[Composition] = None,
) -> Composition:
    if out is not None:
        out._component_tensor[:] = input._component_tensor
        out_component_tensor = out._component_tensor
    else:
        out_component_tensor = input._component_tensor.clone()
    out_residual_tensor = torch.add(
        input._residual_tensor,
        other,
        alpha=alpha,
        out=out._residual_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def ct_add_(
    input: Composition,
    other: Union[Tensor, Number],
    *,
    alpha: Optional[Number] = 1,
) -> Composition:
    out_component_tensor = input._component_tensor
    out_residual_tensor = input._residual_tensor.add_(
        other,
        alpha=alpha,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def tc_add(
    input: Union[Tensor, Number],
    other: Composition,
    *,
    alpha: Optional[Number] = 1,
    out: Optional[Composition] = None,
) -> Composition:
    return ct_add(other, input, alpha=alpha, out=out)


def cc_sub(
    input: Composition,
    other: Composition,
    *,
    alpha: Optional[Number] = 1,
    out: Optional[Composition] = None,
) -> Composition:
    if input.numc() != other.numc():
        raise component_num_error(input.numc(), other.numc())
    out_component_tensor = torch.sub(
        input._component_tensor,
        other._component_tensor,
        alpha=alpha,
        out=out._component_tensor if out is not None else None,
    )
    out_residual_tensor = torch.sub(
        input._residual_tensor,
        other._residual_tensor,
        alpha=alpha,
        out=out.out_residual_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def cc_sub_(
    input: Composition,
    other: Composition,
    *,
    alpha: Optional[Number] = 1,
) -> Composition:
    if input.numc() != other.numc():
        raise component_num_error(input.numc(), other.numc())
    out_component_tensor = input._component_tensor.sub_(
        other._component_tensor,
        alpha=alpha,
    )
    out_residual_tensor = input._residual_tensor.sub_(
        other._residual_tensor,
        alpha=alpha,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def ct_sub(
    input: Composition,
    other: Union[Tensor, Number],
    *,
    alpha: Optional[Number] = 1,
    out: Optional[Composition] = None,
) -> Composition:
    if out is not None:
        out._component_tensor[:] = input._component_tensor
        out_component_tensor = out._component_tensor
    else:
        out_component_tensor = input._component_tensor.clone()
    out_residual_tensor = torch.sub(
        input._residual_tensor,
        other,
        alpha=alpha,
        out=out._residual_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def ct_sub_(
    input: Composition,
    other: Union[Tensor, Number],
    *,
    alpha: Optional[Number] = 1,
) -> Composition:
    out_component_tensor = input._component_tensor
    out_residual_tensor = input._residual_tensor.sub_(
        other,
        alpha=alpha,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def tc_sub(
    input: Union[Tensor, Number],
    other: Composition,
    *,
    alpha: Optional[Number] = 1,
    out: Optional[Composition] = None,
) -> Composition:
    if out is not None:
        out._component_tensor[:] = other._component_tensor
        out_component_tensor = out._component_tensor
    else:
        out_component_tensor = input._component_tensor.clone()
    out_residual_tensor = torch.sub(
        input,
        other._residual_tensor,
        alpha=alpha,
        out=out._residual_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def ct_mul(
    input: Composition,
    other: Union[Tensor, Number],
    *,
    out: Optional[Composition] = None,
) -> Composition:
    if isinstance(other, Tensor) and other.dim() > input.dim():
        # handle broadcast
        new_size = (input.numc(),) + (1,) * (other.dim() - input.dim()) + input.size()
        out_component_tensor = torch.mul(
            input._component_tensor.view(new_size),
            other,
            out=out._component_tensor if out is not None else None,
        )
    else:
        out_component_tensor = torch.mul(
            input._component_tensor,
            other,
            out=out._component_tensor if out is not None else None,
        )
    out_residual_tensor = torch.mul(
        input._residual_tensor,
        other,
        out=out._residual_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def ct_mul_(input: Composition, other: Union[Tensor, Number]) -> Composition:
    if isinstance(other, Tensor) and other.dim() > input.dim():
        # handle broadcast
        new_size = (input.numc(),) + (1,) * (other.dim() - input.dim()) + input.size()
        input._component_tensor = input._component_tensor.view(new_size).mul_(other)
    else:
        input._component_tensor = input._component_tensor.mul_(other)
    input._residual_tensor = input._residual_tensor.mul_(other)
    return input


def tc_mul(
    input: Union[Tensor, Number],
    other: Composition,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    return ct_mul(other, input, out=out)


def ct_div(
    input: Composition,
    other: Union[Tensor, Number],
    *,
    rounding_mode: Optional[str] = None,
    out: Optional[Composition] = None,
) -> Composition:
    if isinstance(other, Tensor) and other.dim() > input.dim():
        # handle broadcast
        new_size = (input.numc(),) + (1,) * (other.dim() - input.dim()) + input.size()
        out_component_tensor = torch.div(
            input._component_tensor.view(new_size),
            other,
            rounding_mode=rounding_mode,
            out=out._component_tensor if out is not None else None,
        )
    else:
        out_component_tensor = torch.div(
            input._component_tensor,
            other,
            rounding_mode=rounding_mode,
            out=out._component_tensor if out is not None else None,
        )
    out_residual_tensor = torch.div(
        input._residual_tensor,
        other,
        rounding_mode=rounding_mode,
        out=out._residual_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def ct_div_(
    input: Composition,
    other: Union[Tensor, Number],
    *,
    rounding_mode: Optional[str] = None,
) -> Composition:
    if isinstance(other, Tensor) and other.dim() > input.dim():
        # handle broadcast
        new_size = (input.numc(),) + (1,) * (other.dim() - input.dim()) + input.size()
        input._component_tensor = input._component_tensor.view(new_size).div_(
            other, rounding_mode=rounding_mode
        )
    else:
        input._component_tensor = input._component_tensor.div_(
            other,
            rounding_mode == rounding_mode,
        )
    input._residual_tensor = input._residual_tensor.div_(
        other,
        rounding_mode=rounding_mode,
    )
    return input


def ct_matmul(
    input: Composition, other: Tensor, *, out: Optional[Composition] = None
) -> Composition:
    out_component_tensor = torch.matmul(
        input._component_tensor,
        other,
        out=out._component_tensor if out is not None else None,
    )
    out_residual_tensor = torch.matmul(
        input._residual_tensor,
        other,
        out=out._residual_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def tc_matmul(
    input: Tensor, other: Composition, *, out: Optional[Composition] = None
) -> Composition:
    if other.dim() == 1:
        # if the component_tensor's ndim is 2, the component dim
        # will be incorrectly included in the multiplication
        out_component_tensor = torch.matmul(
            input,
            other._component_tensor.unsqueeze(-1),
            out=out._component_tensor.unsqueeze_(-1) if out is not None else None,
        )
        out_component_tensor.squeeze_(-1)
        if out is not None:
            out._component_tensor.squeeze_(-1)
    else:
        out_component_tensor = torch.matmul(
            input,
            other._component_tensor,
            out=out._component_tensor if out is not None else None,
        )
    out_residual_tensor = torch.matmul(
        input,
        other._residual_tensor,
        out=out._residual_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def ct_mv(
    input: Composition,
    vec: Tensor,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    out_residual_tensor = torch.mv(
        input._residual_tensor,
        vec,
        out=out._residual_tensor if out is not None else None,
    )
    out_component_tensor = torch.matmul(
        input._component_tensor,
        vec,
        out=out._component_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def tc_mv(
    input: Tensor,
    vec: Composition,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    out_residual_tensor = torch.mv(
        input,
        vec._residual_tensor,
        out=out._residual_tensor if out is not None else None,
    )
    out_component_tensor = torch.matmul(
        input,
        vec._component_tensor.unsqueeze(-1),
        out=out._component_tensor.unsqueeze_(-1) if out is not None else None,
    )
    out_component_tensor.squeeze_(-1)
    if out is not None:
        out._component_tensor.squeeze_(-1)
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def ct_mm(
    input: Composition,
    mat2: Tensor,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    out_residual_tensor = torch.mm(
        input._residual_tensor,
        mat2,
        out=out._residual_tensor if out is not None else None,
    )
    out_component_tensor = torch.matmul(
        input._component_tensor,
        mat2,
        out=out._component_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def tc_mm(
    input: Tensor,
    mat2: Composition,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    out_residual_tensor = torch.mm(
        input,
        mat2._residual_tensor,
        out=out._residual_tensor if out is not None else None,
    )
    out_component_tensor = torch.matmul(
        input,
        mat2._component_tensor,
        out=out._component_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def ct_bmm(
    input: Composition,
    mat2: Tensor,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    out_residual_tensor = torch.bmm(
        input._residual_tensor,
        mat2,
        out=out._residual_tensor if out is not None else None,
    )
    out_component_tensor = torch.matmul(
        input._component_tensor,
        mat2,
        out=out._component_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def tc_bmm(
    input: Tensor,
    mat2: Composition,
    *,
    out: Optional[Composition] = None,
) -> Composition:
    out_residual_tensor = torch.bmm(
        input,
        mat2._residual_tensor,
        out=out._residual_tensor if out is not None else None,
    )
    out_component_tensor = torch.matmul(
        input,
        mat2._component_tensor,
        out=out._component_tensor if out is not None else None,
    )
    return pydec.as_composition(out_component_tensor, out_residual_tensor)



from typing import Type, Union, Optional
import torch
from torch.types import _int, _float


def args_error(
    funcion_name: str, args: Union[list, tuple], kwargs: dict = {}
) -> TypeError:
    arg_types = tuple(type(arg).__name__ for arg in args)
    kwarg_types = tuple(
        f"{key}={type(value).__name__}" for key, value in kwargs.items()
    )
    type_str = ", ".join(arg_types + kwarg_types)
    return TypeError(
        f"{funcion_name}() received an invalid combination of arguments - got ({type_str})"
    )


def arg_value_error(msg: str) -> RuntimeError:
    return RuntimeError(msg)


def size_error(
    s1: torch.Size, s2: torch.Size, t1_name="tensor1", t2_name="tensor2"
) -> RuntimeError:
    return RuntimeError(
        f"The size of {t1_name} [{s1}] doesn't match the size of {t2_name} [{s2}]"
    )


def component_num_error(
    num1: _int, num2: _int, c1_name="composition1", c2_name="composition2"
) -> RuntimeError:
    return RuntimeError(
        f"The component number of {c1_name} ({num1}) doesn't match the component number of {c2_name} ({num2})"
    )


def unsupported_operand_error(
    operation_name: str, operand_type1: Type, operand_type2: Type
) -> TypeError:
    return TypeError(
        f"unsupported operand type(s) for {operation_name}: '{operand_type1.__name__}' and '{operand_type2.__name__}'"
    )


def overflow_error(error: _float, error_bound=1e-2) -> RuntimeError:
    return RuntimeError(
        f"Error overflow in checking, error {error} out of bound {error_bound}"
    )


def none_decomposition_func_error(func_name: Optional[str] = None) -> RuntimeError:
    return RuntimeError(
        f"Cannot find the currently used decomposition function with key name ({func_name})"
    )


# TODO: deprecated
def args_error_not_in_tracing(*args, **kwargs):
    error_msg = (
        "received Composition type argument but not in the tracing mode - got ({}).\nTo fix this error: "
        "(1) input tensor instead if you do not want to do tracing or (2) set tracing enabled by Tracer.trace() or pydec.set_tracing_enabled()"
    )
    arg_types = tuple(type(arg).__name__ for arg in args)
    kwarg_types = tuple(
        f"{key}={type(value).__name__}" for key, value in kwargs.items()
    )
    type_str = ", ".join(arg_types + kwarg_types)
    return RuntimeError(error_msg.format(type_str))



from .states import (
    register_decomposition_func,
    set_decomposition_func,
    get_decomposition_name,
    get_decomposition_func,
    using_decomposition_func,
    set_decomposition_args,
    get_decomposition_args,
    using_decomposition_args,
)
from .ops import *

from .algorithms import *

"""
Initialization
"""
try:
    set_decomposition_func("affine")
except ValueError:
    from . import states

    set_decomposition_func(
        states._DECOMPOSITION_FUNC_REGISTRY.keys().__iter__().__next__()
    )



import pydec
from pydec import Composition
import torch
import pytest

torch.manual_seed(114514)


def init_composition(size, c_num=3, requires_grad=False):
    c = pydec.zeros(size, c_num, dtype=torch.float)
    for i in range(c_num):
        c()[i] = torch.randn(size, requires_grad=requires_grad)
    return c


class TestIndexing:
    @classmethod
    def init_composition(cls) -> Composition:
        size = (3, 4)
        c = pydec.zeros(size, c_num=4)
        return c

    def test1(self):
        c = TestIndexing.init_composition()
        c_ = c[None]
        assert c_.numc() == c.numc()
        assert c_.size() == (1,) + c.size()

        c[None] = 1
        assert torch.all(c._residual_tensor == 0)
        assert torch.all(c._component_tensor == 1)

    def test2(self):
        c = TestIndexing.init_composition()
        c0 = c[0]
        assert isinstance(c0, Composition)
        assert c0.numc() == c.numc()
        assert c0.size() == c.size()[1:]
        c02 = c[0:2]
        assert isinstance(c02, Composition)
        assert c02.numc() == c.numc()
        assert c02.size() == (2,) + c.size()[1:]
        c02_ = c[:, 0:2]
        assert c02_.numc() == c.numc()
        assert c02_.size() == c.size()[:1] + (2,)
        c0202 = c[:2, :2]
        assert c0202.numc() == c.numc()
        assert c0202.size() == (2, 2)

    def test3(self):
        c = TestIndexing.init_composition()
        c_ = c[:, None]
        assert c_.size() == c.size()[:1] + (1,) + c.size()[1:]
        index_list = [0, 2]
        c_ = c[index_list]
        assert isinstance(c_, Composition)
        assert c_.numc() == c.numc()
        assert c_.size() == (2,) + c.size()[1:]
        c_ = c[index_list, index_list]
        assert c_.c_size() == (c.numc(), 2)

        with pytest.raises(IndexError):
            c_ = c[index_list, index_list, index_list]

        with pytest.raises(IndexError):
            c_ = c[:, index_list, index_list]


class TestIndexingWithSyntacticSugar:
    def test1(self):
        c = TestIndexing.init_composition()
        with pytest.raises(RuntimeError):
            c()[None]
        with pytest.raises(RuntimeError):
            c()[None] = 1

    def test2(self):
        c = TestIndexing.init_composition()
        c0 = c()[0]
        assert isinstance(c0, torch.Tensor)
        assert c0.size() == c.size()
        c02 = c()[0:2]
        assert isinstance(c02, Composition)
        assert c02.numc() == 2
        assert c02.size() == c.size()
        c02_ = c()[:, 0:2]
        assert c02_.numc() == c.numc()
        assert c02_.size() == (2,) + c.size()[1:]
        c0202 = c()[:2, :2]
        assert c0202.numc() == 2
        assert c0202.size() == (2,) + c.size()[1:]

    def test3(self):
        c = TestIndexing.init_composition()
        c_ = c()[:, None]
        assert c_.size() == (1,) + c.size()
        index_list = [0, 2]
        c_ = c()[index_list]
        assert isinstance(c_, Composition)
        assert c_.numc() == 2
        assert c_.size() == c.size()
        c_ = c()[index_list, index_list]
        assert c_.c_size() == (2,) + c.c_size()[2:]

        c_ = c()[index_list, index_list, index_list]
        assert c_.c_size() == (2,)

        c_ = c()[:, index_list, index_list]
        assert c_.c_size() == (4, 2)


class Test_C_AccessingIndexing:
    @pydec.enable_c_accessing()
    def test1(self):
        c = TestIndexing.init_composition()
        with pytest.raises(RuntimeError):
            c[None]
        with pytest.raises(RuntimeError):
            c[None] = 1

    @pydec.enable_c_accessing()
    def test2(self):
        c = TestIndexing.init_composition()
        c0 = c[0]
        assert isinstance(c0, torch.Tensor)
        assert c0.size() == c.size()
        c02 = c[0:2]
        assert isinstance(c02, Composition)
        assert c02.numc() == 2
        assert c02.size() == c.size()
        c02_ = c[:, 0:2]
        assert c02_.numc() == c.numc()
        assert c02_.size() == (2,) + c.size()[1:]
        c0202 = c[:2, :2]
        assert c0202.numc() == 2
        assert c0202.size() == (2,) + c.size()[1:]

    @pydec.enable_c_accessing()
    def test3(self):
        c = TestIndexing.init_composition()
        c_ = c[:, None]
        assert c_.size() == (1,) + c.size()
        index_list = [0, 2]
        c_ = c[index_list]
        assert isinstance(c_, Composition)
        assert c_.numc() == 2
        assert c_.size() == c.size()
        c_ = c[index_list, index_list]
        assert c_.c_size() == (2,) + c.c_size()[2:]

        c_ = c[index_list, index_list, index_list]
        assert c_.c_size() == (2,)

        c_ = c[:, index_list, index_list]
        assert c_.c_size() == (4, 2)


class TestView:
    c = init_composition((2, 3))

    def test_view1(self):
        assert self.c.view(torch.float16)._component_tensor.dtype == torch.float16

        assert self.c.view(dtype=torch.float16)._component_tensor.dtype == torch.float16

    def test_view2(self):
        assert self.c.view((3, 2)).size() == (3, 2)
        assert self.c.view(size=(3, 2)).size() == (3, 2)
        assert self.c.view(3, 2).size() == (3, 2)


class TestReshape:
    c = init_composition((2, 3))

    def test1(self):
        assert self.c.reshape((3, 2)).size() == (3, 2)
        assert self.c.reshape(shape=(3, 2)).size() == (3, 2)
        assert self.c.reshape(3, 2).size() == (3, 2)


class TestPlus:
    c = init_composition((2, 3))

    def test1(self):
        from pydec import decOVF

        self.c[:] = 1.5
        decOVF.set_decomposition_func("abs_affine")
        c = self.c + 3
        assert (self.c._residual_tensor + 3 == c._residual_tensor).all()

        # set_decomposition_func("norm_decomposition")
        # c = self.c + 3
        # assert (self.c._residual_tensor + 3 == c._residual_tensor).all()

        decOVF.set_decomposition_func("hybrid_affine")
        c = self.c + 3
        assert (self.c._residual_tensor + 3 == c._residual_tensor).all()

        # set_decomposition_func("sign_decomposition")
        # c = self.c + 3
        # assert (self.c._residual_tensor + 3 == c._residual_tensor).all()

        # set_decomposition_func("hybrid_decomposition_threshold")
        # c = self.c + 3
        # assert (self.c._residual_tensor + 3 == c._residual_tensor).all()

        with decOVF.using_decomposition_func("none"):
            c = self.c + 3
            assert (self.c._component_tensor == c._component_tensor).all()
            assert (self.c._residual_tensor + 3 == c._residual_tensor).all()

    def test2(self):
        c = init_composition((2, 3))
        out = c + self.c
        assert torch.all(
            self.c._component_tensor + c._component_tensor == out._component_tensor
        )



import pydec
from ._composition_test import init_composition
from pydec import decOVF


class TestBiasArgs:
    c = init_composition((2, 3))

    def test_set_args1(self):
        args = {"eps": 1e-2, "arg1": "foo"}
        decOVF.set_decomposition_args(**args)
        bias_args = decOVF.get_decomposition_args()
        for k, v in args.items():
            assert bias_args[k] == v

        args = {"eps": 1e-3, "arg2": None}
        decOVF.set_decomposition_args(**args)
        bias_args = decOVF.get_decomposition_args()
        assert bias_args["arg1"] == "foo"
        for k, v in args.items():
            assert bias_args[k] == v

    def test_set_args2(self):
        args = {"eps": 1e-2, "arg1": "foo"}
        decOVF.set_decomposition_args(update=False, **args)
        bias_args = decOVF.get_decomposition_args()
        assert len(bias_args) == len(args)
        for k, v in args.items():
            assert bias_args[k] == v

        args = {"eps": 1e-3, "arg2": None}
        decOVF.set_decomposition_args(update=False, **args)
        bias_args = decOVF.get_decomposition_args()
        assert len(bias_args) == len(args)
        for k, v in args.items():
            assert bias_args[k] == v

    def test_using_args1(self):
        args = {"threshold": 0.8}
        decOVF.set_decomposition_func("hybrid_affine")
        assert "threshold" not in decOVF.get_decomposition_args()
        with decOVF.using_decomposition_args(**args):
            assert decOVF.get_decomposition_args()["threshold"] == args["threshold"]
            c = self.c + 3

        assert "threshold" not in decOVF.get_decomposition_args()
